import json
from collections import OrderedDict
from itertools import cycle

import numpy as np
import pytest
import zhinst.utils as zi_utils
from zhinst.core import compile_seqc

from zhinst.toolkit.driver.nodes.awg import CommandTableNode, Waveforms


@pytest.fixture()
def waveform_descriptors_json(data_dir):
    json_path = data_dir / "waveform_descriptors.json"
    with json_path.open("r", encoding="UTF-8") as file:
        return file.read()


def test_enable_sequencer(mock_connection, shfsg):
    shfsg.sgchannels[0].awg.enable_sequencer(single=True)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/sgchannels/0/awg/single", 1
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/sgchannels/0/awg/enable", 1
    )

    shfsg.sgchannels[0].awg.enable_sequencer(single=False)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/sgchannels/0/awg/single", 0
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/sgchannels/0/awg/enable", 1
    )


def test_wait_done(mock_connection, shfsg):
    single = 0
    enable = iter([])

    def get_int_side_effect(node):
        if node.upper() == "/DEV1234/SGCHANNELS/0/AWG/SINGLE":
            return single
        if node.upper() == "/DEV1234/SGCHANNELS/0/AWG/ENABLE":
            return next(enable)

    mock_connection.return_value.getInt.side_effect = get_int_side_effect

    # if not single mode this function throws a RuntimeError
    with pytest.raises(RuntimeError) as e_info:
        shfsg.sgchannels[0].awg.wait_done()

    # already finished
    single = 1
    enable = iter([0] * 2)
    shfsg.sgchannels[0].awg.wait_done()
    # finishes in time
    single = 1
    enable = iter([1] * 3 + [0] * 2)
    shfsg.sgchannels[0].awg.wait_done()
    # don't finish
    single = 1
    enable = cycle([1])
    with pytest.raises(TimeoutError) as e_info:
        shfsg.sgchannels[0].awg.wait_done(timeout=0.1)


def test_load_sequencer_program(mock_connection, shfsg):

    # everything ok
    elf, info_original = compile_seqc("setTrigger(1);", "SHFSG8", [], 0)
    info = shfsg.sgchannels[0].awg.load_sequencer_program("setTrigger(1);")
    mock_connection.return_value.set.assert_called_once()
    assert (
        mock_connection.return_value.set.call_args[0][0]
        == "/dev1234/sgchannels/0/awg/elf/data"
    )
    assert mock_connection.return_value.set.call_args[0][1] == elf
    assert info == info_original

    # Compiler error
    with pytest.raises(RuntimeError) as e_info:
        shfsg.sgchannels[0].awg.load_sequencer_program("Hello")

    # Empty string sequencer program
    with pytest.raises(RuntimeError):
        shfsg.sgchannels[0].awg.load_sequencer_program("")
    with pytest.raises(RuntimeError):
        shfsg.sgchannels[0].awg.load_sequencer_program(None)

    # Upload error
    mock_connection.return_value.set.side_effect = RuntimeError()
    mock_connection.return_value.setVector.side_effect = RuntimeError()
    with pytest.raises(RuntimeError):
        elf, info_original = shfsg.sgchannels[0].awg.load_sequencer_program(
            "setTrigger(1);"
        )


def test_load_sequencer_program_qc(mock_connection, shfqc):
    elf, info_original = compile_seqc("setTrigger(1);", "SHFQC", [], 0, sequencer="sg")
    info = shfqc.sgchannels[0].awg.load_sequencer_program("setTrigger(1);")
    mock_connection.return_value.set.assert_called_once()
    assert (
        mock_connection.return_value.set.call_args[0][0]
        == "/dev1234/sgchannels/0/awg/elf/data"
    )
    assert mock_connection.return_value.set.call_args[0][1] == elf
    assert info == info_original


def test_command_table(shfsg):
    assert isinstance(shfsg.sgchannels[0].awg.commandtable, CommandTableNode)
    assert shfsg.sgchannels[0].awg.commandtable.raw_tree == shfsg.sgchannels[
        0
    ].awg.raw_tree + ("commandtable",)


def test_write_to_waveform_memory(waveform_descriptors_json, mock_connection, shfsg):
    mock_connection.return_value.get.return_value = OrderedDict(
        [
            (
                "/dev12044/sgchannels/0/awg/waveform/descriptors",
                [
                    {
                        "timestamp": 1158178198389432,
                        "flags": 0,
                        "vector": waveform_descriptors_json,
                    }
                ],
            )
        ]
    )
    waveforms = Waveforms()
    wave1 = 1.0 * np.ones(1008)
    wave2 = -1.0 * np.ones(1008)
    marker = np.zeros(1008)
    waveforms[0] = (wave1, wave2)
    waveforms[1] = (wave1, wave2, marker)
    shfsg.sgchannels[0].awg.write_to_waveform_memory(waveforms)
    assert (
        mock_connection.return_value.set.call_args[0][0][0][0]
        == "/dev1234/sgchannels/0/awg/waveform/waves/0"
    )
    assert (
        mock_connection.return_value.set.call_args[0][0][1][0]
        == "/dev1234/sgchannels/0/awg/waveform/waves/1"
    )
    assert all(
        mock_connection.return_value.set.call_args[0][0][0][1]
        == waveforms.get_raw_vector(0)
    )
    assert all(
        mock_connection.return_value.set.call_args[0][0][1][1]
        == waveforms.get_raw_vector(1)
    )

    shfsg.sgchannels[0].awg.write_to_waveform_memory(waveforms, [0])
    assert (
        mock_connection.return_value.set.call_args[0][0][0][0]
        == "/dev1234/sgchannels/0/awg/waveform/waves/0"
    )
    assert len(mock_connection.return_value.set.call_args[0][0]) == 1

    # existing transaction
    with shfsg.sgchannels[0].awg.root.set_transaction():
        shfsg.sgchannels[0].awg.dio.highbits(0)
        shfsg.sgchannels[0].awg.write_to_waveform_memory(waveforms)

    mock_connection.return_value.set.call_args[0][0][0][
        0
    ] == "/dev1234/sgchannels/0/awg/dio/highbits"
    assert len(mock_connection.return_value.set.call_args[0][0]) == 3

    # to big index
    waveforms[10] = (wave1, wave2)
    with pytest.raises(IndexError) as e_info:
        shfsg.sgchannels[0].awg.write_to_waveform_memory(waveforms)
    shfsg.sgchannels[0].awg.write_to_waveform_memory(waveforms, validate=False)
    assert (
        mock_connection.return_value.set.call_args[0][0][2][0]
        == "/dev1234/sgchannels/0/awg/waveform/waves/10"
    )
    del waveforms[10]
    # assign to filler
    waveforms[2] = (wave1, wave2)
    with pytest.raises(RuntimeError) as e_info:
        shfsg.sgchannels[0].awg.write_to_waveform_memory(waveforms)
    shfsg.sgchannels[0].awg.write_to_waveform_memory(waveforms, validate=False)
    assert (
        mock_connection.return_value.set.call_args[0][0][2][0]
        == "/dev1234/sgchannels/0/awg/waveform/waves/2"
    )


def test_read_from_waveform_memory(waveform_descriptors_json, mock_connection, shfsg):
    waveform_descriptiors = json.loads(waveform_descriptors_json)

    single_wave_result = []

    def get_side_effect(nodes, **kwargs):
        if "/dev1234/sgchannels/0/awg/waveform/waves/" in nodes.lower():
            if nodes[-1] == "*":
                return OrderedDict(
                    [
                        (
                            "/dev1234/sgchannels/0/awg/waveform/descriptors",
                            [
                                {
                                    "timestamp": 1158178198389432,
                                    "flags": 0,
                                    "vector": json.dumps(waveform_descriptiors),
                                }
                            ],
                        ),
                        (
                            "/dev1234/sgchannels/0/awg/waveform/waves/0",
                            [
                                {
                                    "timestamp": 338544371667920,
                                    "flags": 0,
                                    "vector": zi_utils.convert_awg_waveform(
                                        np.ones(1008), -np.ones(1008), np.ones(1008)
                                    ),
                                }
                            ],
                        ),
                        (
                            "/dev1234/sgchannels/0/awg/waveform/waves/1",
                            [
                                {
                                    "timestamp": 338544371667920,
                                    "flags": 0,
                                    "vector": [],
                                }
                            ],
                        ),
                        (
                            "/dev1234/sgchannels/0/awg/waveform/waves/2",
                            [
                                {
                                    "timestamp": 338544371667920,
                                    "flags": 0,
                                    "vector": [],
                                }
                            ],
                        ),
                    ]
                )
            else:
                return OrderedDict(
                    [
                        (
                            "/dev1234/sgchannels/0/awg/waveform/descriptors",
                            [
                                {
                                    "timestamp": 1158178198389432,
                                    "flags": 0,
                                    "vector": json.dumps(waveform_descriptiors),
                                }
                            ],
                        ),
                        (
                            f"/dev1234/sgchannels/0/awg/waveform/waves/{nodes[-1]}",
                            [
                                {
                                    "timestamp": 338544371667920,
                                    "flags": 0,
                                    "vector": single_wave_result,
                                }
                            ],
                        ),
                    ]
                )
        raise RuntimeError()

    mock_connection.return_value.get.side_effect = get_side_effect
    waveforms = shfsg.sgchannels[0].awg.read_from_waveform_memory()
    mock_connection.return_value.get.assert_called_with(
        "/dev1234/sgchannels/0/awg/waveform/descriptors,/dev1234/sgchannels/0/awg/waveform/waves/*",
        settingsonly=False,
        flat=True,
    )
    assert all(waveforms[0][0] == np.ones(1008))
    assert all(waveforms[0][1] == -np.ones(1008))
    assert all(waveforms[0][2] == np.ones(1008))
    assert all(waveforms[1][0] == np.ones(0))
    assert all(waveforms[1][1] == np.ones(0))
    assert all(waveforms[1][2] == np.ones(0))

    # single Node Access
    single_wave_result = zi_utils.convert_awg_waveform(
        np.ones(1008), -np.ones(1008), np.ones(1008)
    )
    waveforms = shfsg.sgchannels[0].awg.read_from_waveform_memory([0])
    mock_connection.return_value.get.assert_called_with(
        "/dev1234/sgchannels/0/awg/waveform/descriptors,/dev1234/sgchannels/0/awg/waveform/waves/0",
        settingsonly=False,
        flat=True,
    )
    assert len(waveforms) == 1
    assert all(waveforms[0][0] == np.ones(1008))
    assert all(waveforms[0][1] == -np.ones(1008))
    assert all(waveforms[0][2] == np.ones(1008))

    single_wave_result = zi_utils.convert_awg_waveform(
        np.ones(1008), -np.ones(1008), None
    )
    waveform_descriptiors["waveforms"][1]["marker_bits"] = "0;0"
    waveforms = shfsg.sgchannels[0].awg.read_from_waveform_memory([1])
    assert len(waveforms) == 1
    assert all(waveforms[1][0] == np.ones(1008))
    assert all(waveforms[1][1] == -np.ones(1008))
    assert waveforms[1][2] == None

    single_wave_result = zi_utils.convert_awg_waveform(
        np.ones(1008), None, np.ones(1008)
    )
    waveform_descriptiors["waveforms"][1]["channels"] = "1"
    waveform_descriptiors["waveforms"][1]["marker_bits"] = "1;0"
    waveforms = shfsg.sgchannels[0].awg.read_from_waveform_memory([1])
    assert len(waveforms) == 1
    assert all(waveforms[1][0] == np.ones(1008))
    assert waveforms[1][1] == None
    assert all(waveforms[1][2] == np.ones(1008))

    single_wave_result = zi_utils.convert_awg_waveform(np.ones(1008), None, None)
    waveform_descriptiors["waveforms"][1]["channels"] = "1"
    waveform_descriptiors["waveforms"][1]["marker_bits"] = "0;0"
    waveforms = shfsg.sgchannels[0].awg.read_from_waveform_memory([1])
    assert len(waveforms) == 1
    assert all(waveforms[1][0] == np.ones(1008))
    assert waveforms[1][1] == None
    assert waveforms[1][2] == None



"""Helper functions used in toolkit."""
import typing as t
from contextlib import contextmanager
from functools import lru_cache

# TypedDict is available in the typing module since 3.8
# Ift we only support 3.8 we should switch to t.TypedDict
from typing_extensions import TypedDict

T = t.TypeVar("T")

_NodeInfo = TypedDict(
    "_NodeInfo",
    {
        "Node": str,
        "Description": str,
        "Properties": str,
        "Type": str,
        "Unit": str,
        "Options": t.Dict[str, str],
    },
)
NodeDoc = t.Dict[str, _NodeInfo]


def lazy_property(property_function: t.Callable[..., T]) -> property:
    """Alternative for functools.lazy_property.

    functools.lazy_property is only available since python 3.8.
    Should be replaced with functools.lazy_property once no version below
    python 3.8 is supported.

    Args:
        property_function (Callable): property function

    Returns
        Return value of the property function

    """
    return property(lru_cache()(property_function))


@contextmanager
def create_or_append_set_transaction(nodetree) -> t.Generator[None, None, None]:
    """Context manager for a transactional set.

    In contrast to the set_transaction from the nodetree this function only
    creates a new transaction if no other is in progress.
    Should only be called withing the toolkit code.

    Warning:
        This function will silently fail if the existing transaction is exited
        before this function finishes.

    Warning:
        The set is always performed as deep set if called on device nodes.

    Examples:
        >>> with nodetree.set_transaction():
                nodetree.test[0].a(1)
                with create_or_append_set_transaction(nodetree):
                    nodetree.test[1].a(2)
                nodetree.test[2].a(2)
    """
    if not nodetree.transaction.in_progress():
        with nodetree.set_transaction():
            yield
    else:
        yield



from pathlib import Path
import tempfile
from unittest.mock import Mock, patch

from .. import zhinst_toolkit_symlink


@patch("builtins.print")
@patch("scripts.zhinst_toolkit_symlink.os")
def test_create_symlink(mock_os, mock_print):
    with tempfile.TemporaryDirectory() as tmpdirname:
        src = Path(tmpdirname) / "foo/bar"
        src.mkdir(parents=True, exist_ok=True)
        dest = Path(tmpdirname) / "tester"
        mock_os.symlink = Mock()
        zhinst_toolkit_symlink.create_symlink(src, dest)
        mock_os.symlink.assert_called_with(src, dest)
        mock_print.assert_called_with(
            f"Symlink created. Source: {src}, Destination: {dest}"
        )


@patch("builtins.print")
def test_create_symlink_exists(mock_print):
    with tempfile.TemporaryDirectory() as tmpdirname:
        src = Path(tmpdirname) / "foo/bar"
        src.mkdir(parents=True, exist_ok=True)
        dest = Path(tmpdirname) / "foo"
        dest.mkdir(parents=True, exist_ok=True)
        zhinst_toolkit_symlink.create_symlink(src, dest)
        mock_print.assert_called_with(f"Destination: {dest} already exists.")



from unittest.mock import patch

import pytest

from zhinst.toolkit import SHFQAChannelMode
from zhinst.toolkit.driver.devices.shfqa import Generator, QAChannel, Readout, SHFScope


def test_repr(shfqa):
    assert repr(shfqa) == "SHFQA(SHFQA4,DEV1234)"


def test_factory_reset(shfqa):
    # factory reset not yet implemented
    with pytest.warns(RuntimeWarning) as record:
        shfqa.factory_reset()


def test_start_continuous_sw_trigger(mock_connection, shfqa):
    with patch("zhinst.toolkit.driver.devices.shfqa.utils", autospec=True) as utils:
        shfqa.start_continuous_sw_trigger(num_triggers=10, wait_time=20.0)
        utils.start_continuous_sw_trigger.assert_called_once_with(
            mock_connection.return_value, "DEV1234", num_triggers=10, wait_time=20.0
        )


def test_max_qubits_per_channel(shfqa):
    with patch("zhinst.toolkit.driver.devices.shfqa.utils", autospec=True) as utils:
        shfqa.max_qubits_per_channel
        # cached property -> second call should not call device_utils
        shfqa.max_qubits_per_channel
        utils.max_qubits_per_channel.assert_called_once()


def test_qachannels(shfqa):
    assert len(shfqa.qachannels) == 4
    assert isinstance(shfqa.qachannels[0], QAChannel)


def test_scopes(shfqa):
    assert len(shfqa.scopes) == 1
    assert isinstance(shfqa.scopes[0], SHFScope)


def test_qa_configure_channel(mock_connection, shfqa):
    with patch("zhinst.toolkit.driver.devices.shfqa.utils", autospec=True) as utils:
        shfqa.qachannels[0].configure_channel(
            input_range=10,
            output_range=20,
            center_frequency=30.0,
            mode=SHFQAChannelMode.READOUT,
        )
        utils.configure_channel.assert_called_once_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            input_range=10,
            output_range=20,
            center_frequency=30.0,
            mode="readout",
        )


def test_qa_generator(shfqa, data_dir, mock_connection):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    assert isinstance(shfqa.qachannels[0].generator, Generator)
    assert shfqa.qachannels[0].generator.root == shfqa.root
    assert shfqa.qachannels[0].generator.raw_tree == shfqa.raw_tree + (
        "qachannels",
        "0",
        "generator",
    )


def test_qa_readout(shfqa):
    assert isinstance(shfqa.qachannels[0].readout, Readout)
    assert shfqa.qachannels[0].readout.root == shfqa.root
    assert shfqa.qachannels[0].readout.raw_tree == shfqa.raw_tree + (
        "qachannels",
        "0",
        "readout",
    )



import numpy as np
import pytest

from zhinst.toolkit.driver.modules.base_module import BaseModule


@pytest.fixture()
def base_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    yield BaseModule(mock_connection.return_value.awgModule(), session)


def test_repr(base_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(base_module)


def test_raw_module(base_module, mock_connection):
    assert base_module.raw_module == mock_connection.return_value.awgModule()


def test_device(zi_devices_json, base_module, mock_connection):
    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices":
            return zi_devices_json
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    # get
    mock_connection.return_value.awgModule.return_value.getString.return_value = ""
    result = base_module.device()
    assert result == ""

    mock_connection.return_value.awgModule.return_value.getString.return_value = (
        "dev1234"
    )
    result = base_module.device()
    assert result.serial == "dev1234"

    # set
    base_module.device("dev1634")
    mock_connection.return_value.awgModule.return_value.set.assert_called_with(
        "/device", "dev1634"
    )

    base_module.device(result)
    mock_connection.return_value.awgModule.return_value.set.assert_called_with(
        "/device", "dev1234"
    )


def test_wait_done(base_module, mock_connection):
    module_mock = mock_connection.return_value.awgModule.return_value
    module_mock.finished.return_value = 1
    base_module.wait_done()

    module_mock.finished.return_value = 0
    module_mock.progress.return_value = np.array([0.1])
    with pytest.raises(TimeoutError) as e_info:
        base_module.wait_done(timeout=0.01)


def test_subscribe(base_module, mock_connection):
    module_mock = mock_connection.return_value.awgModule.return_value
    base_module.subscribe(base_module.test)
    module_mock.subscribe.assert_called_with("/test")


def test_unsubscribe(base_module, mock_connection):
    module_mock = mock_connection.return_value.awgModule.return_value
    base_module.unsubscribe(base_module.test)
    module_mock.unsubscribe.assert_called_with("/test")


def test_execute(base_module, mock_connection):
    module_mock = mock_connection.return_value.awgModule.return_value
    base_module.execute()
    module_mock.execute.assert_called_once()



"""Dictionary like waveform representation."""
import typing as t
import warnings
from collections.abc import MutableMapping
from enum import IntFlag

import numpy as np
from zhinst.utils import convert_awg_waveform, parse_awg_waveform

_Waveform = t.Tuple[np.ndarray, t.Optional[np.ndarray], t.Optional[np.ndarray]]


class OutputType(IntFlag):
    """Waveform output type.

    OUT1: Enables the output 1 for the respective wave.
    OUT2: Enables the output 2 for the respective wave.

    .. versionadded:: 0.3.5
    """

    OUT1 = 1
    OUT2 = 2


class Wave(np.ndarray):
    """Numpy array subclass containing additional waveform metadata.

    This class takes a standard ndarray that already exists, casts as Wave
    type, and adds the following extra attributes/metadata:
    * name
    * output

    The additional metadata is only used for the sequencer code generation.

    (Based on https://numpy.org/doc/stable/user/basics.subclassing.html)

    Args:
        input_array: existing ndarray
        name: optional name of the waveform in the sequencer code snippet.
        output: optional output configuration for the waveform in the
                sequencer code snippet.

    .. versionadded:: 0.3.5
    """

    def __new__(
        cls,
        input_array,
        name: t.Optional[str] = None,
        output: t.Optional[OutputType] = None,
    ) -> "Wave":
        """Casts an existing ndarray to a Wave type.

        Args:
            input_array: existing ndarray
            name: optional name of the waveform in the sequencer code snippet.
            output: optional output configuration for the waveform in the
                    sequencer code snippet.

        Returns:
            Array as Wave object.
        """
        obj = np.asarray(input_array).view(cls)
        obj.name = name
        obj.output = output
        return obj

    def __array_finalize__(self, obj: t.Optional[np.ndarray]) -> None:
        if obj is None:
            return
        self.name = getattr(obj, "name", None)
        self.output = getattr(obj, "output", None)


class Waveforms(MutableMapping):
    """Waveform dictionary.

    The key specifies the slot of the waveform on the device.
    The value is a the waveform itself, represented by a tuple
    (wave1, wave2, marker).

    The value tuple(wave1, wave2=None, marker=None) consists of the following parts:
        * wave1 (array): Array with data of waveform 1.
        * wave2 (array): Array with data of waveform 2.
        * markers (array): Array with marker data.

    A helper function exist called `assign_waveform` which provides an easy way
    of assigning waveforms to slots. But one can also use the direct dictionary
    access:

    >>> wave = 1.0 * np.ones(1008)
    >>> markers = np.zeros(1008)
    >>> waveforms = Waveforms()
    >>> waveforms.assign_waveform(0, wave)
    >>> waveforms.assign_waveform(1, wave, -wave)
    >>> waveforms.assign_waveform(2, wave, -wave, markers)
    >>> waveforms.assign_waveform(3, wave, markers=markers)
    >>> waveforms[4] = (wave,)
    >>> waveforms[5] = (wave, -wave)
    >>> waveforms[6] = (wave, -wave, markers)
    >>> waveforms[7] = (wave, None, markers)

    The arrays can be provided as arrays of integer, float. The first wave also
    can be of type complex. In that case the second waveform must be `None`.

    Depending on the target format the function `get_raw_vector` converts the
    waves into the following format:

    * native AWG waveform format (interleaved waves and markers as uint16) that
      can be uploaded to the AWG waveform nodes. In case the first wave is of
      type complex the imaginary part is treated as the second wave.
    * complex waveform format that can be uploaded to the generator waveform
      nodes (does not support markers). In case two real waveforms have been
      specified they are combined into a single complex waveform, where the
      imaginary part defined by the second wave.
    """

    def __init__(self):
        self._waveforms = {}

    def __getitem__(self, slot: int) -> _Waveform:
        return self._waveforms[slot]

    def __setitem__(self, slot: int, value: t.Union[np.ndarray, _Waveform]):
        if isinstance(value, np.ndarray):
            self._set_waveform(slot, (value, None, None))
        else:
            self._set_waveform(slot, value)

    def __delitem__(self, slot: int):
        del self._waveforms[slot]

    def __iter__(self):
        return iter(self._waveforms)

    def __len__(self):
        return len(self._waveforms)

    def assign_waveform(
        self,
        slot: int,
        wave1: np.ndarray,
        wave2: t.Optional[np.ndarray] = None,
        markers: t.Optional[np.ndarray] = None,
    ) -> None:
        """Assigns a waveform to a slot.

        Args:
            slot: slot number
            wave1: Array with data of waveform 1.
            wave2: Array with data of waveform 2. (default = None)
            markers: Array with marker data. (default = None)
        """
        self._set_waveform(slot, (wave1, wave2, markers))

    def assign_native_awg_waveform(
        self,
        slot: int,
        raw_waveform: np.ndarray,
        channels: int = 1,
        markers_present: bool = False,
    ) -> None:
        """Assigns a native AWG waveform to a slot.

        Native AWG waveform = a single waveform (interleaved waves and markers
        as uint16).

        Args:
            slot: slot number
            raw_waveform: native AWG waveform.
            channels: Number of channels present in the wave. (default = 1)
            markers_present: Indicates if markers are interleaved in the wave.
                (default = False)
        """
        wave1, wave2, markers = parse_awg_waveform(
            raw_waveform,
            channels=channels,
            markers_present=markers_present,
        )
        if markers_present and channels == 2:
            self._waveforms[slot] = (wave1, wave2, markers)
        elif channels == 2:
            self._waveforms[slot] = (wave1, wave2, None)
        elif markers_present:
            self._waveforms[slot] = (wave1, None, markers)
        else:
            self._waveforms[slot] = (wave1, None, None)

    def _set_waveform(
        self,
        slot: int,
        value: _Waveform,
    ) -> None:
        """Assigns a tuple of waves to the slot.

        The passed waves are validated against the following requirements:
        * At least one wave must be defined
        * At most three waves are defined
        * The waves must by numpy arrays
        * The waves must have the same length
        * If the first wave is complex teh second wave must be None

        Raises:
            RuntimeError: If the tuple does not comply to the requirements.
        """
        if len(value) < 1 or len(value) > 3:
            raise RuntimeError(
                "Only one(complex) or two(real) waveforms (plus an optional marker) "
                f"can be specified per waveform. ({len(value)} where specified."
            )
        if (
            not isinstance(value[0], np.ndarray)
            or (
                len(value) > 2
                and value[1] is not None
                and not isinstance(value[1], np.ndarray)
            )
            or (
                len(value) > 3
                and value[1] is not None
                and not isinstance(value[2], np.ndarray)
            )
        ):
            raise RuntimeError("Waveform must be specified as numpy.arrays")
        if len(value) >= 2 and value[1] is not None and len(value[0]) != len(value[1]):
            raise RuntimeError("The two waves must have the same length")
        if len(value) == 3 and value[2] is not None and len(value[0]) != len(value[2]):
            raise RuntimeError(
                "The marker must have the same length than the waveforms"
            )
        if np.iscomplexobj(value[0]) and not (len(value) < 3 or value[1] is None):
            raise RuntimeError(
                "The first waveform is complex therefore only one "
                "waveform can be specified."
            )
        self._waveforms[slot] = tuple(
            w.view(Wave) if w is not None else None for w in value
        ) + (None,) * (3 - len(value))

    def get_raw_vector(
        self,
        slot: int,
        *,
        target_length: t.Optional[int] = None,
        complex_output: bool = False,
    ) -> np.ndarray:
        """Get the raw vector for a slot required by the device.

        Either converts a waveform into the native AWG waveform format that can
        be uploaded to the AWG wave node or converts the waveform into a complex
        waveform that can be uploaded to a generator wave node.
        (complex_output = True).

        Args:
            slot: slot number of the waveform
            target_length: target length of the output waveform. If
                specified the waves and markers will have the specified length
                (e.g. wave1, wave2 and markers are specified the resulting
                waveform will have the length 3*target_length). (default = None)
            complex_output: Flag if the output should be a complex waveform for a
                generator node, instead of of the native AWG format that can
                only be uploaded to an AWG node. (default = False)

        Returns:
            Waveform in the native AWG format or as a complex waveform

        Raises:
            ValueError: The length of the waves does not match the target length.
        """
        waves = self._waveforms[slot]
        wave1 = np.zeros(1) if len(waves[0]) == 0 else waves[0]
        wave2 = np.zeros(1) if waves[1] is not None and len(waves[1]) == 0 else waves[1]
        marker = waves[2]

        if target_length and len(wave1) > target_length:
            raise ValueError(
                f"Waveforms are larger than the target length "
                f"{len(wave1)} > {target_length}."
            )
        if target_length and len(wave1) < target_length:
            raise ValueError(
                f"Waveforms are smaller than the target length "
                f"{len(wave1)} < {target_length}."
            )

        if complex_output and np.iscomplexobj(wave1):
            if wave2 is not None or marker is not None:
                warnings.warn("Complex values do not support markers", RuntimeWarning)
            return wave1
        if complex_output and not np.iscomplexobj(wave1):
            if marker is not None:
                warnings.warn("Complex values do not support markers", RuntimeWarning)
            complex_wave = np.zeros(wave1.shape, dtype=np.complex128)
            complex_wave.real = wave1
            if wave2 is not None:
                complex_wave.imag = wave2
            return complex_wave

        if np.iscomplexobj(wave1):
            marker = wave2 if wave2 is not None else marker
            wave2 = wave1.imag
            wave1 = wave1.real

        return convert_awg_waveform(
            wave1,
            wave2=wave2,
            markers=marker if marker is not None else None,
        )

    def _get_waveform_sequence(self, index: int) -> str:
        """Get sequencer code snippet for a single waveform.

        The sequencer code snippet is generated with the following information:
            * Waveform length
            * Waveform index
            * presence of markers and for which channel
            * Defined names of the waveforms (if set)
            * Defined output configuration (if set)

        Returns:
            Sequencer code snippet.

        .. versionadded:: 0.3.5
        """
        waves = self._waveforms[index]
        wave_length = max(1, waves[0].size)
        w2_present = waves[1] is not None
        marker = waves[2]
        names = [waves[0].name, waves[1].name if waves[1] is not None else None]
        outputs = [waves[0].output, waves[1].output if waves[1] is not None else None]

        if np.iscomplexobj(waves[0]):
            marker = waves[1] if waves[1] is not None else marker
            w2_present = True
            names = names if not names[0] or isinstance(names[0], str) else names[0]
            outputs = (
                outputs
                if not outputs[0] or not isinstance(outputs, t.Iterable)
                else outputs[0]
            )
        marker = None if marker is None else np.unpackbits(marker.astype(np.uint8))

        def marker_to_bool(i: int) -> str:
            return "true" if np.any(marker[7 - i :: 8]) else "false"  # noqa: E203

        def to_wave_str(i: int) -> str:
            if marker is None:
                return f"placeholder({wave_length}, false, false)"
            return (
                f"placeholder({wave_length}, {marker_to_bool(i*2)}, "
                + f"{marker_to_bool(i*2+1)})"
            )

        w1_assign = to_wave_str(0)
        w2_assign = to_wave_str(1) if w2_present else ""
        w2_decl = w1_decl = ""
        if names[0]:
            w1_decl = f"wave {names[0]} = {w1_assign};\n"
            w1_assign = names[0]
        if names[1]:
            w2_decl = f"wave {names[1]} = {w2_assign};\n"
            w2_assign = names[1]
        if outputs[0]:
            if outputs[0] in [OutputType.OUT1, OutputType.OUT2]:
                w1_assign = f"{outputs[0]}, {w1_assign}"
            elif outputs[0] == OutputType.OUT1 | OutputType.OUT2:
                w1_assign = f"1, 2, {w1_assign}"
        if outputs[1]:
            if outputs[1] in [OutputType.OUT1, OutputType.OUT2]:
                w2_assign = f"{outputs[1]}, {w2_assign}"
            elif outputs[1] == OutputType.OUT1 | OutputType.OUT2:
                w2_assign = f"1, 2, {w2_assign}"

        if w2_assign:
            return (
                f"{w1_decl}{w2_decl}assignWaveIndex({w1_assign}, {w2_assign}, {index});"
            )
        return f"{w1_decl}assignWaveIndex({w1_assign}, {index});"

    def get_sequence_snippet(self) -> str:
        """Return a sequencer code snippet for the defined waveforms.

        Based on the defined waveforms and their additional information this
        function generates a sequencer code snippet that can be used to define
        the given waveforms. The following information will be used:

            * Waveform length
            * Waveform index
            * presence of markers and for which channel
            * Defined names of the waveforms (if set)
            * Defined output configuration (if set)

        Example:
            >>> waveform = Waveform()
            >>> waveform.assign_waveform(
                    0,
                    wave1=Wave(
                        np.ones(1008),
                        name="w1",
                        output=OutputType.OUT1 | OutputType.OUT2
                    ),
                    wave2=Wave(
                        -np.ones(1008),
                        name="w2",
                        output=OutputType.OUT2),
                    markers=15 * np.ones(1008),
                )
            >>> waveform.get_sequence_snippet()
            wave w1 = placeholder(1008, true, true);
            wave w2 = placeholder(1008, true, true);
            assignWaveIndex(1, 2, w1, 2, w2, 0);

        Returns:
            Sequencer Code snippet.

        .. versionadded:: 0.3.5
        """
        return "\n".join(
            [
                self._get_waveform_sequence(slot)
                for slot in sorted(self._waveforms.keys())
            ]
        )



"""Implements Connection wrapper around a native python dictionary."""
import fnmatch
import json
import re
import typing as t
from collections import OrderedDict

from numpy import array

from zhinst.toolkit.nodetree.helper import NodeDoc


class ConnectionDict:
    """Connection wrapper around a dictionary.

    The ``NodeTree`` expects a connection that complies to the
    protocol :class:`nodetree.Connection`. In order to also support raw
    dictionaries this class wraps around a python dictionary and exposes the
    required protocol.

    Args:
        data: Dictionary raw path: value
        json_info: JSON information for each path (path: info)
    """

    def __init__(self, data: t.Dict[str, t.Any], json_info: NodeDoc):
        super().__init__()
        self._values = data
        self.json_info = json_info

    def listNodesJSON(self, path: str, *args, **kwargs) -> str:
        """Returns a list of nodes with description found at the specified path."""
        if path == "*":
            return json.dumps(self.json_info)
        json_info = {}
        for node, info in self.json_info.items():
            if fnmatch.fnmatchcase(node, path + "*"):
                json_info[node] = info
        return json.dumps(json_info)

    def get(self, path: str, *args, **kwargs) -> t.Any:
        """Mirrors the behavior of zhinst.core get command."""
        nodes_raw = fnmatch.filter(self._values.keys(), path)
        return_value = OrderedDict()
        for node in nodes_raw:
            return_value[node] = array([self._values[node]])
        return return_value

    def getInt(self, path: str) -> int:
        """Mirrors the behavior of zhinst.core getInt command."""
        try:
            return int(self._values[path])
        except TypeError:
            if self._values[path] is None:
                return 0
            raise

    def getDouble(self, path: str) -> float:
        """Mirrors the behavior of zhinst.core getDouble command."""
        return float(self._values[path])

    def getString(self, path: str) -> str:
        """Mirrors the behavior of zhinst.core getDouble command."""
        return str(self._values[path])

    def _parse_input_value(self, path: str, value: t.Any):
        if isinstance(value, str):
            option_map = {}
            for key, option in self.json_info[path].get("Options", {}).items():
                node_options = re.findall(r'"(.+?)"[,:]+', option)
                option_map.update({x: int(key) for x in node_options})
            return option_map.get(value, value)
        return value

    def set(
        self,
        path: t.Union[str, t.List[t.Tuple[str, t.Any]]],
        value: t.Any = None,
        **kwargs,
    ) -> None:
        """Mirrors the behavior of zhinst.core set command."""
        if isinstance(path, str):
            self._values[path] = self._parse_input_value(path, value)
        else:
            for node, node_value in path:
                self._values[node] = self._parse_input_value(node, node_value)

    def setVector(self, path: str, value: t.Any = None) -> None:
        """Mirrors the behavior of zhinst.core setVector command."""
        self.set(path, value)

    def subscribe(self, path: str) -> None:
        """Mirrors the behavior of zhinst.core subscribe command."""
        raise RuntimeError("Can not subscribe within the SHFQA_Sweeper")

    def unsubscribe(self, path: str) -> None:
        """Mirrors the behavior of zhinst.core unsubscribe command."""
        raise RuntimeError("Can not subscribe within the SHFQA_Sweeper")



import numpy as np
import pytest

from zhinst.toolkit.driver.parsers import Parse


class TestParsers:
    def test_set_true_false(self):
        assert Parse.from_bool(True) == 1
        assert Parse.from_bool(False) == 0

    def test_get_true_false(self):
        assert Parse.to_bool(1) == True
        assert Parse.to_bool(0) == False

    def test_phase(self):
        assert Parse.phase(90) == 90
        assert Parse.phase(450) == 90
        assert Parse.phase(870) == 150

    def test_greater_equal(self, caplog):
        assert Parse.greater_equal(90, 0) == 90
        assert Parse.greater_equal(0, 0) == 0
        assert Parse.greater_equal(-20, 10) == 10
        assert len(caplog.records) == 1
        assert Parse.greater_equal(9.999999999999, 10) == 10
        assert len(caplog.records) == 2

    def test_smaller_equal(self, caplog):
        assert Parse.smaller_equal(-90, 0) == -90
        assert Parse.smaller_equal(0, 0) == 0
        assert Parse.smaller_equal(20, 10) == 10
        assert len(caplog.records) == 1
        assert Parse.smaller_equal(200000000000, 10) == 10
        assert len(caplog.records) == 2

    def test_multiple_of(self, caplog):
        assert Parse.multiple_of(10, 6, "nearest") == 12
        assert len(caplog.records) == 1
        assert Parse.multiple_of(10, 6, "down") == 6
        assert len(caplog.records) == 2
        assert Parse.multiple_of(12, 6, "down") == 12
        assert len(caplog.records) == 2



"""Base Instrument Driver.

Natively works with all device types and provides the basic functionality like
the device specific nodetree.
"""
import copy
import json
import logging
import re
import typing as t
import warnings
from functools import lru_cache
from pathlib import Path

from zhinst.utils._version import version as utils_version_str
from zhinst.core import __version__ as zhinst_version_str

from zhinst.toolkit._min_version import _MIN_DEVICE_UTILS_VERSION, _MIN_LABONE_VERSION
from zhinst.toolkit.driver.parsers import node_parser
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import lazy_property

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from zhinst.toolkit.session import Session


class BaseInstrument(Node):
    """Generic toolkit driver for a Zurich Instrument device.

    All device specific class are derived from this class.
    It exposes the nodetree and also implements common functions valid for all
    devices.
    It also can be used directly, e.g. for instrument types that have no special
    class in toolkit.

    Args:
        serial: Serial number of the device, e.g. *'dev12000'*.
            The serial number can be found on the back panel of the instrument.
        device_type: Type of the device.
        session: Session to the Data Server
    """

    def __init__(
        self,
        serial: str,
        device_type: str,
        session: "Session",
    ):
        self._serial = serial
        self._device_type = device_type
        self._session = session
        try:
            self._options = session.daq_server.getString(f"/{serial}/features/options")
        except RuntimeError:
            self._options = ""

        # HF2 does not support listNodesJSON so we have the information hardcoded
        # (the node of HF2 will not change any more so this is safe)
        preloaded_json = None
        if "HF2" in self._device_type:
            preloaded_json = self._load_preloaded_json(
                Path(__file__).parent / "../../resources/nodedoc_hf2.json"
            )

        nodetree = NodeTree(
            self._session.daq_server,
            prefix_hide=self._serial,
            list_nodes=[f"/{self._serial}/*"],
            preloaded_json=preloaded_json,
        )
        # Add predefined parseres (in node_parser) to nodetree nodes
        nodetree.update_nodes(
            node_parser.get(self.__class__.__name__, {}), raise_for_invalid_node=False
        )

        super().__init__(nodetree, tuple())

    def __repr__(self):
        options = f"({self._options})" if self._options else ""
        options = options.replace("\n", ",")
        return str(
            f"{self.__class__.__name__}({self._device_type}" f"{options},{self.serial})"
        )

    def factory_reset(self, *, deep: bool = True) -> None:
        """Load the factory default settings.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after loading the factory preset (default: True).
        """
        self.system.preset.load(1, deep=deep)
        logger.info(f"Factory preset is loaded to device {self.serial.upper()}.")

    @staticmethod
    def _version_string_to_tuple(version: str) -> t.Tuple[int, int, int]:
        """Converts a version string into a version tuple.

        Args:
            version: Version

        Returns:
            Version as a tuple of ints
        """
        result = [0] * 3
        for i, value in enumerate(version.split(".")):
            try:
                result[i] = int(value)
            except ValueError:
                if i < 3:  # ignore dev verisons
                    result[i] = 0
        return result[0], result[1], result[2]

    @staticmethod
    def _check_python_versions(
        zi_python_version: t.Tuple[int, int, int],
        zi_utils_version: t.Tuple[int, int, int],
    ) -> None:
        """Check if the minimum required zhinst packages are installed.

        Checks if all zhinst packages that toolkit require have the minimum
        required version installed.

        Args:
            zi_python_version: zhinst.core package version
            zi_utils_version: zhinst.utils package version

        Raises:
            RuntimeError: If the zhinst.core version does not match the
                minimum requirements for zhinst.toolkit
            RuntimeError: If the zhinst.utils version does not match the
                minimum requirements for zhinst.toolkit
        """
        if zi_python_version < BaseInstrument._version_string_to_tuple(
            _MIN_LABONE_VERSION
        ):
            raise RuntimeError(
                "zhinst.core version does not match the minimum required version "
                f"for zhinst.toolkit {zi_python_version} < {_MIN_LABONE_VERSION}. "
                "Use `pip install --upgrade zhinst` to get the latest version."
            )
        if zi_utils_version < BaseInstrument._version_string_to_tuple(
            _MIN_DEVICE_UTILS_VERSION
        ):
            raise RuntimeError(
                "zhinst.utils version does not match the minimum required "
                f"version for zhinst.toolkit {zi_utils_version} < "
                f"{_MIN_DEVICE_UTILS_VERSION}. Use `pip install "
                "--upgrade zhinst.utils` to get the latest version."
            )

    @staticmethod
    def _check_labone_version(
        zi_python_version: t.Tuple[int, int, int],
        labone_version: t.Tuple[int, int, int],
    ) -> None:
        """Check that the LabOne version matches the zhinst version.

        Args:
            zi_python_version: zhinst.core package version
            labone_version: LabOne DataServer version

        Raises:
            RuntimeError: If the zhinst.core version does not match the
                version of the connected LabOne DataServer.
        """
        if labone_version[:2] < zi_python_version[:2]:
            raise RuntimeError(
                "The LabOne version is smaller than the zhinst.core version. "
                f"{labone_version} < {zi_python_version}. "
                "Please install the latest/matching LabOne version from "
                "https://www.zhinst.com/support/download-center."
            )
        if labone_version[:2] > zi_python_version[:2]:
            raise RuntimeError(
                "the zhinst.core version is smaller than the LabOne version "
                f"{zi_python_version} < {labone_version}. "
                "Please install the latest/matching version from pypi.org."
            )
        if labone_version[-1] != zi_python_version[-1]:
            warnings.warn(
                "The patch version of zhinst.core and the LabOne DataServer "
                f"mismatch {labone_version[-1]} ! {zi_python_version[-1]}.",
                RuntimeWarning,
                stacklevel=2,
            )

    def _check_firmware_update_status(self) -> None:
        """Check if the firmware matches LabOne version.

        Raises:
            ConnectionError: If the device is currently updating
            RuntimeError: If the firmware revision does not match to th
                version of the connected LabOne DataServer.
        """
        device_info = json.loads(self._session.daq_server.getString("/zi/devices"))[
            self.serial.upper()
        ]
        status_flag = device_info["STATUSFLAGS"]
        if status_flag & 1 << 8:
            raise ConnectionError(
                "The device is currently updating please try again after the update "
                "process is complete"
            )
        if status_flag & 1 << 4 or status_flag & 1 << 5:
            raise RuntimeError(
                "The Firmware does not match the LabOne version. "
                "Please update the firmware (e.g. in the LabOne UI)"
            )
        if status_flag & 1 << 6 or status_flag & 1 << 7:
            raise RuntimeError(
                "The Firmware does not match the LabOne version. "
                "Please update LabOne to the latest version from "
                "https://www.zhinst.com/support/download-center."
            )

    def check_compatibility(self) -> None:
        """Check if the software stack is compatible.

        Only if all versions and revisions of the software stack match stability
        can be ensured. The following criteria are checked:

            * minimum required zhinst-utils package is installed
            * minimum required zhinst-core package is installed
            * zhinst package matches the LabOne Data Server version
            * firmware revision matches the LabOne Data Server version

        Raises:
            ConnectionError: If the device is currently updating
            RuntimeError: If one of the above mentioned criterion is not
                fulfilled
        """
        self._check_python_versions(
            self._version_string_to_tuple(zhinst_version_str),
            self._version_string_to_tuple(utils_version_str),
        )
        labone_version_str = self._session.about.version()
        labone_revision_str = str(self._session.about.revision())[4:]
        labone_version = self._version_string_to_tuple(
            labone_version_str + "." + labone_revision_str
        )
        self._check_labone_version(
            self._version_string_to_tuple(zhinst_version_str), labone_version
        )
        self._check_firmware_update_status()

    @lru_cache()
    def get_streamingnodes(self) -> t.List[Node]:
        """Create a list with all streaming nodes available.

        Returns:
            Available streaming node.
        """
        streaming_nodes = []
        for node, info in self:
            if "Stream" in info.get("Properties"):
                streaming_nodes.append(node)
        return streaming_nodes

    def _load_preloaded_json(self, filename: Path) -> t.Optional[dict]:
        """Load a preloaded json and match the existing nodes.

        Args:
            Filename for the preloaded json.

        Returns:
            Loaded JSON if the file exists.
        """
        if not filename.is_file():
            return None
        raw_file = filename.open("r").read()

        raw_file = raw_file.replace("devxxxx", self.serial.lower())
        raw_file = raw_file.replace("DEVXXXX", self.serial.upper())
        json_raw = json.loads(raw_file)

        existing_nodes = self._session.daq_server.listNodes(
            f"/{self.serial}/*", recursive=True, leavesonly=True
        )

        preloaded_json = {}
        for node in existing_nodes:
            node_name = re.sub(r"(?<!values)\/[0-9]*?$", "/n", node.lower())
            node_name = re.sub(r"\/[0-9]*?\/", "/n/", node_name)
            json_element = copy.deepcopy(json_raw.get(node_name))
            if json_element:
                json_element["Node"] = node.upper()
                preloaded_json[node.lower()] = json_element
            elif not node.startswith("/zi/"):
                print(f"unkown node {node}")

        return preloaded_json

    def set_transaction(self) -> t.ContextManager:
        """Context manager for a transactional set.

        Can be used as a context in a with statement and bundles all node set
        commands into a single transaction. This reduces the network overhead
        and often increases the speed.

        Within the with block a set commands to a node will be buffered
        and bundled into a single command at the end automatically.
        (All other operations, e.g. getting the value of a node, will not be
        affected)

        Warning:
            The set is always performed as deep set if called on device nodes.

        Examples:
            >>> with device.set_transaction():
                    device.test[0].a(1)
                    device.test[1].a(2)
        """
        return self._root.set_transaction()

    @property
    def serial(self) -> str:
        """Instrument specific serial."""
        return self._serial

    @property
    def device_type(self) -> str:
        """Type of the instrument (e.g. MFLI)."""
        return self._device_type

    @lazy_property
    def device_options(self) -> str:
        """Enabled options of the instrument."""
        return self.features.options()



import json
import pytest
from pathlib import Path

from zhinst.toolkit.command_table import CommandTable


CT_SCHEMA_22_02 = Path(__file__).parent / "data/command_table_schema_22_02.json"
CT_SCHEMA_22_08 = Path(__file__).parent / "data/command_table_schema_22_08.json"


@pytest.fixture(scope="function")
def command_table_completed():
    """L1 22.02"""
    with open(Path(__file__).parent / "data/command_table_completed.json") as f:
        data = json.load(f)
    yield data


@pytest.fixture(scope="module")
def command_table_schema_22_02():
    """L1 22.02"""
    with open(CT_SCHEMA_22_02) as f:
        data = json.load(f)
    return data


@pytest.fixture(scope="module")
def command_table_schema_22_08():
    """L1 22.08"""
    with open(CT_SCHEMA_22_08) as f:
        data = json.load(f)
    return data


@pytest.fixture(scope="module", params=[CT_SCHEMA_22_02, CT_SCHEMA_22_08])
def command_table_schema(request):
    with open(request.param) as f:
        data = json.load(f)
    return data


@pytest.fixture
def command_table(command_table_schema):
    yield CommandTable(command_table_schema)



from unittest.mock import patch

import numpy as np
import pytest

from zhinst.toolkit.waveform import Waveforms


@pytest.fixture()
def readout(shfqa):
    yield shfqa.qachannels[0].readout


def test_configure_result_logger(mock_connection, readout):
    with patch("zhinst.toolkit.driver.nodes.readout.utils", autospec=True) as utils:
        readout.configure_result_logger(result_source="test", result_length=10)
        utils.configure_result_logger_for_readout.assert_called_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            result_source="test",
            result_length=10,
            num_averages=1,
            averaging_mode=0,
        )
        readout.configure_result_logger(
            result_source="test2", result_length=0, num_averages=2, averaging_mode=1
        )
        utils.configure_result_logger_for_readout.assert_called_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            result_source="test2",
            result_length=0,
            num_averages=2,
            averaging_mode=1,
        )


def test_run(mock_connection, readout):
    with patch("zhinst.toolkit.driver.nodes.readout.utils", autospec=True) as utils:
        readout.run()
        utils.enable_result_logger.assert_called_with(
            mock_connection.return_value,
            "DEV1234",
            0,
            mode="readout",
        )


def test_stop(mock_connection, readout):
    # already disabled
    mock_connection.return_value.getInt.return_value = 0
    readout.stop()
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/qachannels/0/readout/result/enable", False
    )
    # never disabled
    mock_connection.return_value.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e_info:
        readout.stop(timeout=0.5)


def test_wait_done(mock_connection, readout):
    # already disabled
    mock_connection.return_value.getInt.return_value = 0
    readout.wait_done()
    # never disabled
    mock_connection.return_value.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e_info:
        readout.wait_done(timeout=0.5)


def test_read(mock_connection, readout):
    with patch("zhinst.toolkit.driver.nodes.readout.utils", autospec=True) as utils:
        readout.read()
        utils.get_result_logger_data.assert_called_with(
            mock_connection.return_value, "DEV1234", 0, mode="readout", timeout=10
        )
        readout.read(timeout=1)
        utils.get_result_logger_data.assert_called_with(
            mock_connection.return_value, "DEV1234", 0, mode="readout", timeout=1
        )


def test_write_integration_weights(mock_connection, readout):
    waveforms = Waveforms()
    waveforms[0] = np.ones(1000)
    waveforms[1] = np.ones(1000, dtype=np.complex128)
    waveforms[2] = (np.ones(1000), np.ones(1000))

    waveforms_long = Waveforms()
    waveforms_long[1000] = np.zeros(1000)
    with patch("zhinst.toolkit.driver.nodes.readout.utils", autospec=True) as utils:
        readout.write_integration_weights(waveforms)
        complex_wave = np.empty(1000, dtype=np.complex128)
        complex_wave.real = np.ones(1000)
        complex_wave.imag = np.ones(1000)
        assert (
            utils.configure_weighted_integration.call_args[0][0]
            == mock_connection.return_value
        )
        assert utils.configure_weighted_integration.call_args[0][1] == "DEV1234"
        assert utils.configure_weighted_integration.call_args[0][2] == 0
        assert np.allclose(
            utils.configure_weighted_integration.call_args[1]["weights"][1],
            np.ones(1000, dtype=np.complex128),
        )
        assert np.allclose(
            utils.configure_weighted_integration.call_args[1]["weights"][2],
            complex_wave,
        )
        assert (
            utils.configure_weighted_integration.call_args[1]["integration_delay"]
            == 0.0
        )
        assert (
            utils.configure_weighted_integration.call_args[1]["clear_existing"] == True
        )

        readout.write_integration_weights(waveforms, integration_delay=0.5)
        assert (
            utils.configure_weighted_integration.call_args[1]["integration_delay"]
            == 0.5
        )

        readout.write_integration_weights(waveforms, clear_existing=False)
        assert (
            utils.configure_weighted_integration.call_args[1]["clear_existing"] == False
        )

        readout.write_integration_weights({0: np.ones(1000)}, clear_existing=True)
        assert all(
            utils.configure_weighted_integration.call_args[1]["weights"][0]
            == np.ones(1000)
        )

    with pytest.raises(RuntimeError) as e_info:
        readout.write_integration_weights(waveforms_long)


def test_read_integration_weights(mock_connection, readout):

    result = readout.read_integration_weights()
    assert len(result) == 0
    mock_connection.return_value.get.assert_called_once_with(
        "/dev1234/qachannels/0/readout/integration/weights/*/wave",
        settingsonly=False,
        flat=True,
    )

    result = readout.read_integration_weights([0, 3])
    mock_connection.return_value.get.assert_called_with(
        "/dev1234/qachannels/0/readout/integration/weights/0/wave,/dev1234/qachannels/0/readout/integration/weights/3/wave",
        settingsonly=False,
        flat=True,
    )

    mock_connection.return_value.get.return_value = {
        "/dev1234/qachannels/0/readout/integration/weights/0/wave": [
            {"vector": np.ones(1000, dtype=np.complex128)}
        ],
        "/dev1234/qachannels/0/readout/integration/weights/1/wave": [
            {"vector": np.ones(1000, dtype=np.complex128)}
        ],
    }
    result = readout.read_integration_weights()
    assert len(result) == 2
    np.allclose(result[0][0], np.ones(1000, dtype=np.complex128))
    assert not result[0][1]
    assert not result[0][2]
    np.allclose(result[1][0], np.ones(1000, dtype=np.complex128))



from dataclasses import make_dataclass
from unittest.mock import patch

import pytest

from zhinst.toolkit.driver.modules.shfqa_sweeper import SHFQASweeper, SweepConfig


@pytest.fixture()
def mock_shf_sweeper():
    with patch(
        "zhinst.toolkit.driver.modules.shfqa_sweeper.CoreSweeper", autospec=True
    ) as sweeper:
        yield sweeper


@pytest.fixture()
def mock_SweepConfig():
    with patch(
        "zhinst.toolkit.driver.modules.shfqa_sweeper.SweepConfig",
        make_dataclass("Y", fields=[("s", str, 0)]),
    ) as sweeper_config:
        yield sweeper_config


@pytest.fixture()
def mock_TriggerConfig():
    with patch(
        "zhinst.toolkit.driver.modules.shfqa_sweeper.TriggerConfig",
        make_dataclass("Y", fields=[("s", str, 0)]),
    ) as trigger_config:
        yield trigger_config


@pytest.fixture()
def sweeper_module(session, mock_sweeper_daq, mock_shf_sweeper):
    yield SHFQASweeper(session)


def test_repr(sweeper_module):
    assert "SHFQASweeper(DataServerSession(localhost:8004))" in repr(sweeper_module)


def test_missing_node(mock_connection, mock_shf_sweeper, mock_sweeper_daq, session):
    with patch(
        "zhinst.toolkit.driver.modules.shfqa_sweeper.SweepConfig",
        make_dataclass("Y", fields=[("s", str, 0)], bases=(SweepConfig,)),
    ) as sweeper_config:
        sweeper_module = SHFQASweeper(session)
        assert sweeper_module.sweep.s() == 0


def test_device(
    mock_connection, sweeper_module, session, mock_shf_sweeper, mock_sweeper_daq
):
    assert sweeper_module.device() == ""

    sweeper_module.device("dev1234")
    mock_shf_sweeper.assert_called_with(mock_sweeper_daq(), "dev1234")
    assert sweeper_module.device() == "dev1234"

    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    assert sweeper_module.device() == session.devices["dev1234"]


def test_update_settings(
    mock_connection, sweeper_module, mock_shf_sweeper, mock_sweeper_daq
):
    assert not sweeper_module.envelope.enable()
    mock_shf_sweeper.assert_called_with(sweeper_module._daq_server, "")
    # device needs to be set first
    with pytest.raises(RuntimeError) as e_info:
        sweeper_module._update_settings()

    sweeper_module.device("dev1234")
    mock_shf_sweeper.assert_called_with(mock_sweeper_daq(), "dev1234")
    sweeper_module._update_settings()
    mock_shf_sweeper.return_value.configure.assert_called_once()
    assert "sweep_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert "rf_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert "avg_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert "trig_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert "envelope_config" not in mock_shf_sweeper.return_value.configure.call_args[1]

    sweeper_module.sweep.mapping("log")
    sweeper_module.rf.output_range(1)
    sweeper_module.average.integration_time(1)
    sweeper_module.trigger.level(0.6)
    sweeper_module._update_settings()
    assert "sweep_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert (
        mock_shf_sweeper.return_value.configure.call_args[1]["sweep_config"].mapping
        == sweeper_module.sweep.mapping.node_info.enum.log.name
    )
    assert "rf_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert (
        mock_shf_sweeper.return_value.configure.call_args[1]["rf_config"].output_range
        == 1
    )
    assert "avg_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert (
        mock_shf_sweeper.return_value.configure.call_args[1][
            "avg_config"
        ].integration_time
        == 1
    )
    assert "trig_config" in mock_shf_sweeper.return_value.configure.call_args[1]
    assert (
        mock_shf_sweeper.return_value.configure.call_args[1]["trig_config"].level == 0.6
    )
    assert "envelope_config" not in mock_shf_sweeper.return_value.configure.call_args[1]

    # envelope
    sweeper_module.envelope.enable(True)
    sweeper_module._update_settings()
    envelope = mock_shf_sweeper.return_value.configure.call_args[-1]["envelope_config"]
    assert envelope.delay == 0.0


def test_update_settings_broken(
    mock_TriggerConfig,
    mock_SweepConfig,
    mock_connection,
    sweeper_module,
    mock_shf_sweeper,
    caplog,
):

    sweeper_module.device("dev1234")
    sweeper_module._update_settings()
    assert (
        "imp50 setting is no longer available in the shf_sweeper class."
        in caplog.messages
    )
    assert (
        "use_sequencer setting is no longer available in the shf_sweeper class."
        in caplog.messages
    )
    assert (
        "source setting is no longer available in the shf_sweeper class."
        in caplog.messages
    )
    assert (
        "force_sw_trigger setting is no longer available in the shf_sweeper."
        in caplog.messages
    )


def test_run(sweeper_module, mock_shf_sweeper):

    sweeper_module.device("dev1234")
    sweeper_module.run()
    mock_shf_sweeper.return_value.configure.assert_called_once()
    mock_shf_sweeper.return_value.run.assert_called_once()


def test_get_result(sweeper_module, mock_shf_sweeper):

    sweeper_module.device("dev1234")
    sweeper_module.get_result()
    mock_shf_sweeper.return_value.configure.assert_called_once()
    mock_shf_sweeper.return_value.get_result.assert_called_once()


def test_plot(sweeper_module, mock_shf_sweeper):

    sweeper_module.device("dev1234")
    sweeper_module.plot()
    mock_shf_sweeper.return_value.configure.assert_called_once()
    mock_shf_sweeper.return_value.plot.assert_called_once()


def test_get_offset_freq_vector(sweeper_module, mock_shf_sweeper):

    sweeper_module.device("dev1234")
    sweeper_module.get_offset_freq_vector()
    mock_shf_sweeper.return_value.configure.assert_called_once()
    mock_shf_sweeper.return_value.get_offset_freq_vector.assert_called_once()



import json
import operator
from contextlib import _GeneratorContextManager
from unittest.mock import MagicMock, Mock, patch

import pytest

from zhinst.toolkit._min_version import _MIN_DEVICE_UTILS_VERSION, _MIN_LABONE_VERSION
from zhinst.toolkit.driver.devices.base import BaseInstrument


@pytest.fixture()
def base_instrument(mock_connection, session, nodedoc_dev1234_json):
    mock_connection.return_value.listNodesJSON.return_value = nodedoc_dev1234_json
    mock_connection.return_value.getString.return_value = "OptionA"

    yield BaseInstrument("DEV1234", "test_type", session)


def test_basic_setup(mock_connection, base_instrument):
    mock_connection.return_value.listNodesJSON.assert_called_with(
        f"/{base_instrument.serial}/*"
    )

    assert base_instrument.device_type == "test_type"
    assert base_instrument.serial == "DEV1234"
    assert repr(base_instrument) == "BaseInstrument(test_type(OptionA),DEV1234)"


def test_hf2_setup(data_dir, mock_connection, hf2_session):
    list_nodes_path = data_dir / "list_nodes_hf2_dev.txt"
    with list_nodes_path.open("r", encoding="UTF-8") as file:
        nodes_dev = file.read().split("\n")[:-1]
    mock_connection.return_value.listNodes.return_value = nodes_dev
    mock_connection.return_value.getString.return_value = "OptionA"

    instrument = BaseInstrument("DEV1234", "HF2LI", hf2_session)

    assert instrument.device_type == "HF2LI"
    assert instrument.serial == "DEV1234"
    assert repr(instrument) == "BaseInstrument(HF2LI(OptionA),DEV1234)"


def test_factory_reset(base_instrument, mock_connection):
    base_instrument.factory_reset()
    dev_id = base_instrument.serial.lower()
    mock_connection.return_value.syncSetInt.assert_called_once_with(
        f"/{dev_id}/system/preset/load", 1
    )


def test_get_streamingnodes(base_instrument):
    nodes = base_instrument.get_streamingnodes()

    assert base_instrument.dios[0].input in nodes
    assert base_instrument.scopes[0].wave in nodes
    assert base_instrument.demods[0].sample in nodes


def test_set_transaction(base_instrument):
    assert isinstance(base_instrument.set_transaction(), _GeneratorContextManager)

    base_instrument._root = Mock()
    base_instrument.set_transaction()
    base_instrument.root.set_transaction.assert_called_once()


def test_node_access(base_instrument):

    assert base_instrument.demods[0].rate == base_instrument.root.demods[0].rate
    assert base_instrument["demods"][0].rate == base_instrument.root.demods[0].rate

    assert "demods" in base_instrument
    for element in dir(base_instrument.root):
        assert element in dir(base_instrument)


def test_iter_nodetree(base_instrument):
    for res_base, res_nodetree in zip(base_instrument, base_instrument.root):
        assert res_base == res_nodetree


def test_load_preloaded_json(base_instrument, mock_connection, data_dir):

    assert base_instrument._load_preloaded_json(data_dir) == None

    mock_connection.return_value.listNodes.side_effect = [["/dev1234/stats/0/temp"]]
    return_value = base_instrument._load_preloaded_json(
        data_dir / "preloadable_nodetree.json"
    )
    assert len(return_value) == 1
    assert "/dev1234/stats/0/temp" in return_value

    mock_connection.return_value.listNodes.side_effect = [["/dev1234/stats/0/test"]]
    return_value = base_instrument._load_preloaded_json(
        data_dir / "preloadable_nodetree.json"
    )
    assert len(return_value) == 0


def test_check_python_versions():
    sub = lambda x, y: tuple(map(operator.sub, x, y))
    add = lambda x, y: tuple(map(operator.add, x, y))

    labone = BaseInstrument._version_string_to_tuple(_MIN_LABONE_VERSION)
    device_utils = BaseInstrument._version_string_to_tuple(_MIN_DEVICE_UTILS_VERSION)

    # matching version
    BaseInstrument._check_python_versions(labone, device_utils)
    # zhinst.core heigher
    BaseInstrument._check_python_versions(add(labone, (1, 0, 0)), device_utils)
    # zhinst.core lower
    with pytest.raises(RuntimeError):
        BaseInstrument._check_python_versions(sub(labone, (1, 0, 0)), device_utils)
    # utils heigher
    BaseInstrument._check_python_versions(
        labone,
        add(
            device_utils,
            (1, 0, 0),
        ),
    )
    # utils lower
    with pytest.raises(RuntimeError):
        BaseInstrument._check_python_versions(labone, sub(device_utils, (1, 0, 0)))


def test_check_labone_version():

    # matching version
    BaseInstrument._check_labone_version((21, 8, 0), (21, 8, 0))
    # zhinst.core heigher
    with pytest.raises(RuntimeError):
        BaseInstrument._check_labone_version((22, 2, 0), (21, 8, 0))
    # zhinst.core lower
    with pytest.raises(RuntimeError):
        BaseInstrument._check_labone_version((21, 2, 0), (21, 8, 0))
    # patchversion missmatch
    with pytest.warns(RuntimeWarning):
        BaseInstrument._check_labone_version((21, 8, 1), (21, 8, 0))


def test_check_firmware_update_status(base_instrument, mock_connection):
    info = {
        "DEV1111": {},
        "DEV1234": {
            "AVAILABLE": 1,
            "STATUSFLAGS": 0,
        },
        "DEV1235": {},
    }
    mock_connection.return_value.getString.return_value = json.dumps(info)

    # everything ok
    base_instrument._check_firmware_update_status()

    # Upgrade available
    info = {
        "DEV1111": {},
        "DEV1234": {
            "AVAILABLE": 1,
            "STATUSFLAGS": 1 << 4,
        },
        "DEV1235": {},
    }
    mock_connection.return_value.getString.return_value = json.dumps(info)
    with pytest.raises(RuntimeError):
        base_instrument._check_firmware_update_status()
    info = {
        "DEV1111": {},
        "DEV1234": {
            "AVAILABLE": 1,
            "STATUSFLAGS": 1 << 5,
        },
        "DEV1235": {},
    }
    mock_connection.return_value.getString.return_value = json.dumps(info)
    with pytest.raises(RuntimeError):
        base_instrument._check_firmware_update_status()

    # Downgrade available
    info = {
        "DEV1111": {},
        "DEV1234": {
            "AVAILABLE": 1,
            "STATUSFLAGS": 1 << 6,
        },
        "DEV1235": {},
    }
    mock_connection.return_value.getString.return_value = json.dumps(info)
    with pytest.raises(RuntimeError):
        base_instrument._check_firmware_update_status()
    info = {
        "DEV1111": {},
        "DEV1234": {
            "AVAILABLE": 1,
            "STATUSFLAGS": 1 << 7,
        },
        "DEV1235": {},
    }
    mock_connection.return_value.getString.return_value = json.dumps(info)
    with pytest.raises(RuntimeError):
        base_instrument._check_firmware_update_status()

    # Update in progress
    info = {
        "DEV1111": {},
        "DEV1234": {
            "AVAILABLE": 1,
            "STATUSFLAGS": 1 << 8,
        },
        "DEV1235": {},
    }
    mock_connection.return_value.getString.return_value = json.dumps(info)
    with pytest.raises(ConnectionError):
        base_instrument._check_firmware_update_status()


def test_check_compatibility(base_instrument):

    with patch(
        "zhinst.toolkit.driver.devices.base.BaseInstrument._check_python_versions",
        new_callable=MagicMock,
    ) as check_python_versions:
        with patch(
            "zhinst.toolkit.driver.devices.base.BaseInstrument._check_labone_version",
            new_callable=MagicMock,
        ) as check_labone_version:
            with patch(
                "zhinst.toolkit.driver.devices.base.BaseInstrument._check_firmware_update_status",
                new_callable=MagicMock,
            ) as check_firmware_update_status:
                base_instrument.check_compatibility()
                check_python_versions.assert_called_once()
                check_labone_version.assert_called_once()
                check_firmware_update_status.assert_called_once()



"""UHFLI Instrument Driver."""

import logging
import typing as t

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.nodes.awg import AWG
from zhinst.toolkit.nodetree.helper import lazy_property
from zhinst.toolkit.nodetree.node import Node, NodeList

logger = logging.getLogger(__name__)


class UHFLI(BaseInstrument):
    """High-level driver for the Zurich Instruments UHFLI."""

    @lazy_property
    def awgs(self) -> t.Union[t.Sequence[AWG], Node]:
        """A Sequence of AWG Cores.

        Device options requirement(s): AWG
        """
        if "AWG" not in self.features.options():
            logger.error("Missing option: AWG")
            return Node(
                self._root,
                self._tree + ("awgs",),
            )
        return NodeList(
            [
                AWG(
                    self.root,
                    self._tree + ("awgs", str(i)),
                    self._session,
                    self.serial,
                    i,
                    self.device_type,
                )
                for i in range(len(self["awgs"]))
            ],
            self._root,
            self._tree + ("awgs",),
        )



from unittest.mock import patch

import pytest


@pytest.fixture()
def spectroscopy(shfqa):
    yield shfqa.qachannels[0].spectroscopy


@pytest.fixture()
def m_utils():
    with patch(
        "zhinst.toolkit.driver.nodes.spectroscopy.utils", autospec=True
    ) as utils:
        yield utils


def test_configure_result_logger(mock_connection, spectroscopy, m_utils):
    spectroscopy.configure_result_logger(result_length=10)
    m_utils.configure_result_logger_for_spectroscopy.assert_called_with(
        mock_connection.return_value,
        "DEV1234",
        0,
        result_length=10,
        num_averages=1,
        averaging_mode=0,
    )
    spectroscopy.configure_result_logger(
        result_length=0, num_averages=2, averaging_mode=1
    )
    m_utils.configure_result_logger_for_spectroscopy.assert_called_with(
        mock_connection.return_value,
        "DEV1234",
        0,
        result_length=0,
        num_averages=2,
        averaging_mode=1,
    )


def test_run(mock_connection, spectroscopy, m_utils):
    spectroscopy.run()
    m_utils.enable_result_logger.assert_called_with(
        mock_connection.return_value,
        "DEV1234",
        0,
        mode="spectroscopy",
    )


def test_stop(mock_connection, spectroscopy):
    # already disabled
    mock_connection.return_value.getInt.return_value = 0
    spectroscopy.stop()
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/qachannels/0/spectroscopy/result/enable", False
    )
    # never disabled
    mock_connection.return_value.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e_info:
        spectroscopy.stop(timeout=0.5)


def test_wait_done(mock_connection, spectroscopy):
    # already disabled
    mock_connection.return_value.getInt.return_value = 0
    spectroscopy.wait_done()
    # never disabled
    mock_connection.return_value.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e_info:
        spectroscopy.wait_done(timeout=0.5)


def test_read(mock_connection, spectroscopy, m_utils):
    spectroscopy.read()
    m_utils.get_result_logger_data.assert_called_with(
        mock_connection.return_value, "DEV1234", 0, mode="spectroscopy", timeout=10
    )
    spectroscopy.read(timeout=1)
    m_utils.get_result_logger_data.assert_called_with(
        mock_connection.return_value, "DEV1234", 0, mode="spectroscopy", timeout=1
    )



from collections import OrderedDict
from itertools import cycle
from socket import timeout

import numpy as np
import pytest

from zhinst.toolkit.driver.devices.pqsc import PQSC


@pytest.fixture()
def pqsc(data_dir, mock_connection, session):

    json_path = data_dir / "nodedoc_dev1234_pqsc.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.listNodesJSON.return_value = nodes_json
    mock_connection.return_value.getString.return_value = ""

    yield PQSC("DEV1234", "PQSC", session)


def test_repr(pqsc):
    assert repr(pqsc) == "PQSC(PQSC,DEV1234)"


def test_arm(mock_connection, pqsc):
    pqsc.arm()
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/execution/enable", False
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/feedback/registerbank/reset", 1
    )

    pqsc.arm(repetitions=3, holdoff=100.6)
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/execution/enable", False
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/repetitions", 3
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/holdoff", 100.6
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/feedback/registerbank/reset", 1
    )

    pqsc.arm(deep=False, repetitions=5, holdoff=67.4)
    mock_connection.return_value.set.assert_any_call("/dev1234/execution/enable", False)
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/repetitions", 5
    )
    mock_connection.return_value.set.assert_any_call("/dev1234/execution/holdoff", 67.4)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/feedback/registerbank/reset", 1
    )


def test_run_stop(mock_connection, pqsc):
    pqsc.run(deep=False)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/execution/enable", True
    )
    pqsc.run()
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/execution/enable", True
    )

    pqsc.stop(deep=False)
    mock_connection.return_value.set.assert_called_with(
        "/dev1234/execution/enable", False
    )
    pqsc.stop()
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/execution/enable", False
    )
    mock_connection.reset_mock()

    pqsc.arm_and_run(repetitions=5, holdoff=100.8)
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/execution/enable", False
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/repetitions", 5
    )
    mock_connection.return_value.set.assert_any_call(
        "/dev1234/execution/holdoff", 100.8
    )
    mock_connection.return_value.syncSetInt.assert_any_call(
        "/dev1234/feedback/registerbank/reset", 1
    )
    mock_connection.return_value.syncSetInt.assert_called_with(
        "/dev1234/execution/enable", True
    )


def test_wait_done(mock_connection, pqsc):
    mock_connection.return_value.getInt.side_effect = [1, 1, 1, 0, 0, 0, 0, 0]
    pqsc.wait_done()
    mock_connection.return_value.getInt.assert_called_with("/dev1234/execution/enable")

    mock_connection.return_value.getInt.side_effect = None
    mock_connection.return_value.getInt.return_value = 1
    with pytest.raises(TimeoutError) as e_info:
        pqsc.wait_done(timeout=0.2)
    mock_connection.return_value.getInt.assert_called_with("/dev1234/execution/enable")


def test_ref_clock(mock_connection, pqsc):

    status = cycle([0])
    source = 0
    source_actual = 0

    def get_side_effect(path):
        if path == "/dev1234/system/clocks/referenceclock/in/status":
            return next(status)
        if path == "/dev1234/system/clocks/referenceclock/in/source":
            return source
        if path == "/dev1234/system/clocks/referenceclock/in/sourceactual":
            return source_actual
        raise RuntimeError("Invalid Node")

    mock_connection.return_value.getInt.side_effect = get_side_effect

    assert pqsc.check_ref_clock()
    # Locked within time
    status = iter([2] * 2 + [0] * 10)
    assert pqsc.check_ref_clock()
    # Locking error but actual_clock == clock
    status = cycle([1])
    assert not pqsc.check_ref_clock()
    # Locking error and actual_clock != clock => reset clock to internal
    source = 1
    mock_connection.return_value.syncSetString.assert_not_called()
    assert not pqsc.check_ref_clock()
    mock_connection.return_value.syncSetString.assert_called_with(
        "/dev1234/system/clocks/referenceclock/in/source", "internal"
    )

    # timeout
    status = cycle([2])
    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_ref_clock(timeout=0.2)


def test_check_zsync_connection(mock_connection, pqsc):

    cycle([2])

    status0 = 0
    status2 = 0
    status10 = 0

    def getInt_side_effect(path):
        if path == "/dev1234/zsyncs/0/connection/status":
            return status0
        if path == "/dev1234/zsyncs/2/connection/status":
            return status2
        if path == "/dev1234/zsyncs/10/connection/status":
            return status10
        raise RuntimeError("Invalid Node")

    mock_connection.return_value.getInt.side_effect = getInt_side_effect

    # not connected
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection(timeout=0)
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection(2, timeout=0)
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection([0], timeout=0)
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection([0, 2], timeout=0)

    # one connected
    status0 = 2
    assert True == pqsc.check_zsync_connection()
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection(2, timeout=0)
    assert [True] == pqsc.check_zsync_connection([0])
    with pytest.raises(TimeoutError):
        pqsc.check_zsync_connection([0, 2], timeout=0)

    # one connected the other one in error state
    status0 = 2
    status2 = 3
    assert True == pqsc.check_zsync_connection()
    assert False == pqsc.check_zsync_connection(2)
    assert [True] == pqsc.check_zsync_connection([0])
    assert [True, False] == pqsc.check_zsync_connection([0, 2])

    # one connected the other one times out
    status0 = 2
    status2 = 1
    assert True == pqsc.check_zsync_connection()
    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection(2, timeout=0.01)
    assert [True] == pqsc.check_zsync_connection([0])

    with pytest.raises(TimeoutError) as e_info:
        pqsc.check_zsync_connection([0, 2], timeout=0.01)
    assert "2" in str(e_info.value)

    # test two decimal place value
    status10 = 2
    status0 = 2
    status2 = 3
    assert [True] == pqsc.check_zsync_connection([10])
    assert [True, False, True] == pqsc.check_zsync_connection([0, 2, 10])



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))


# -- Project information -----------------------------------------------------

project = "zhinst-toolkit"
copyright = "2020, Zurich Instruments AG"
author = "Zurich Instruments AG"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosectionlabel",
    "sphinx_issues",
    "nbsphinx",
    "nbsphinx_link",
    "IPython.sphinxext.ipython_console_highlighting",
    "m2r2",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
]
add_module_names = False
# Autodoc settings
autodoc_default_options = {"show-inheritance": True}
autodoc_typehints = "both"
autodoc_typehints_format = "short"
autosummary_generate = True
set_type_checking_flag = False

# Sphinx issues
issues_github_path = "zhinst/zhinst-toolkit"

# Make sure the target is unique
autosectionlabel_prefix_document = True
autosectionlabel_maxdepth = 2

nbsphinx_execute = "never"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "**.ipynb_checkpoints"]

from importlib.metadata import version

version = version("zhinst.toolkit")


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = ['zhinst-sphinx-theme/css/custom.css']

html_theme_options = {
    "logo": {
        "text": "zhinst-toolkit",
    }
}

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True

# nbsphinx parameters

nbsphinx_codecell_lexer = "none"
highlight_language = "none"

# Spelling
# sphinxcontrib.spelling configuration file
spelling_word_list_filename='spelling_wordlist.txt'
# Show suggestion in console output
spelling_show_suggestions=False
spelling_exclude_patterns=['examples/*.nblink', 'source/_static/zhinst-sphinx-theme/**/*']



import numpy as np
import time
import numpy as np
import typing as t
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt


def voltage_to_power_dBm(voltage):
    power = 10 * np.log10(np.abs(voltage) ** 2 / 50 * 1000)
    return power


def voltage_to_phase(voltage):
    phase = np.unwrap(np.angle(voltage))
    return phase


def generate_flat_top_gaussian(
    frequencies, pulse_duration, rise_fall_time, sampling_rate, scaling=0.9
):
    """Returns complex flat top Gaussian waveforms modulated with the given frequencies.

    Args:

        frequencies (array): array specifying the modulation frequencies applied to each
                             output wave Gaussian

        pulse_duration (float): total duration of each Gaussian in seconds

        rise_fall_time (float): rise-fall time of each Gaussian edge in seconds

        sampling_rate (float): sampling rate in samples per second based on which to
                               generate the waveforms

        scaling (optional float): scaling factor applied to the generated waveforms (<=1);
                                  use a scaling factor <= 0.9 to avoid overshoots

    Returns:

        pulses (dict): dictionary containing the flat top Gaussians as values

    """
    if scaling > 1:
        raise ValueError(
            "The scaling factor has to be <= 1 to ensure the generated waveforms lie within the \
                unit circle."
        )

    from scipy.signal import gaussian

    rise_fall_len = int(rise_fall_time * sampling_rate)
    pulse_len = int(pulse_duration * sampling_rate)

    std_dev = rise_fall_len // 10

    gauss = gaussian(2 * rise_fall_len, std_dev)
    flat_top_gaussian = np.ones(pulse_len)
    flat_top_gaussian[0:rise_fall_len] = gauss[0:rise_fall_len]
    flat_top_gaussian[-rise_fall_len:] = gauss[-rise_fall_len:]

    flat_top_gaussian *= scaling

    pulses = {}
    time_vec = np.linspace(0, pulse_duration, pulse_len)

    for i, f in enumerate(frequencies):
        pulses[i] = flat_top_gaussian * np.exp(2j * np.pi * f * time_vec)

    return pulses


def set_trigger_loopback(session, device, rate=1000):
    """
    Start a continuous trigger pulse from marker 1 A using the internal loopback to trigger in 1 A
    """

    m_ch = 0
    low_trig = 2
    continuous_trig = 1
    daq = session.daq_server
    daq.syncSetInt(f"/{device.serial}/raw/markers/*/testsource", low_trig)
    daq.setInt(f"/{device.serial}/raw/markers/{m_ch}/testsource", continuous_trig)
    daq.setDouble(f"/{device.serial}/raw/markers/{m_ch}/frequency", rate)
    daq.setInt(f"/{device.serial}/raw/triggers/{m_ch}/loopback", 1)
    time.sleep(0.2)


def clear_trigger_loopback(session, device):
    m_ch = 0
    session.daq_server.setInt(f"/{device.serial}/raw/markers/*/testsource", 0)
    session.daq_server.setInt(f"/{device.serial}/raw/triggers/{m_ch}/loopback", 0)


def run_experiment(device, sgchannel_number, number_of_qubits, reenable=False):
    if reenable:
        with device.set_transaction():
            device.qachannels[0].readout.run()
            device.qachannels[0].generator.enable_sequencer(single=True)
            device.qachannels[0].input.on(1)
            device.qachannels[0].output.on(1)
            for qubit in range(number_of_qubits):
                channel = sgchannel_number[qubit]
                device.sgchannels[channel].awg.enable(1)
                device.sgchannels[channel].output.on(1)
    else:
        with device.set_transaction():
            device.qachannels[0].input.on(1)
            device.qachannels[0].output.on(1)
            for qubit in range(number_of_qubits):
                device.sgchannels[sgchannel_number[qubit]].output.on(1)

    device.start_continuous_sw_trigger(num_triggers=1, wait_time=2e-3)

    readout_results = device.qachannels[0].readout.read(timeout=100)

    with device.set_transaction():
        device.qachannels[0].input.on(0)
        device.qachannels[0].output.on(0)
        for qubit in range(number_of_qubits):
            device.sgchannels[sgchannel_number[qubit]].output.on(0)

    return readout_results


def fit_data(
    x_data,
    y_data,
    function,
    start_param,
    do_plot,
    figsize=10,
    font=10,
    qubit=0,
    x_label="",
    y_label="",
    saveloc="",
):

    pars, cov = curve_fit(
        f=function, xdata=x_data, ydata=y_data, p0=start_param, bounds=(-np.inf, np.inf)
    )
    stdevs = np.sqrt(np.diag(cov))
    residuals = y_data - function(x_data, *pars)

    # plot it
    if do_plot:
        fig4, axs = plt.subplots(2, 1, figsize=figsize)
        fig4.suptitle(f"Qubit {qubit}", fontsize=font)
        axs[0].scatter(x_data, y_data, s=20, color="#00b3b3", label="Data")
        axs[0].plot(
            x_data, function(x_data, *pars), linestyle="--", linewidth=2, color="black"
        )
        axs[0].legend(loc=2, prop={"size": font})

        axs[0].set_xlabel(x_label, fontsize=font)
        axs[0].set_ylabel(y_label, fontsize=font)
        axs[0].tick_params(axis="both", which="major", labelsize=font)

        axs[1].plot(x_data, residuals, ".")
        axs[1].set_xlabel(x_label, fontsize=font)
        axs[1].set_ylabel("residuals", fontsize=font)
        axs[1].tick_params(axis="both", which="major", labelsize=font)

        plt.savefig(saveloc)
        plt.show()

    return pars, stdevs


def amplitude_rabi(amp, Omega, amplitude):
    return -amplitude * np.cos(amp * Omega / 2) + amplitude



import json
from unittest.mock import patch

import pytest

from zhinst.toolkit import Session
from zhinst.toolkit.driver.modules.base_module import BaseModule
from zhinst.toolkit.driver.modules.daq_module import DAQModule
from zhinst.toolkit.driver.modules.shfqa_sweeper import SHFQASweeper
from zhinst.toolkit.driver.modules.sweeper_module import SweeperModule
from zhinst.toolkit.nodetree import Node


def test_setup(mock_connection, session):
    mock_connection.assert_called_once_with("localhost", 8004, 6)
    mock_connection.return_value.listNodesJSON.assert_called_once_with("/zi/*")
    assert repr(session) == "DataServerSession(localhost:8004)"
    assert not session.is_hf2_server
    assert session.server_host == "localhost"
    assert session.server_port == 8004


def test_setup_hf2(mock_connection, hf2_session):
    mock_connection.assert_called_once_with("localhost", 8005, 1)
    mock_connection.return_value.listNodesJSON.assert_not_called()
    assert repr(hf2_session) == "HF2DataServerSession(localhost:8005)"
    assert hf2_session.is_hf2_server


def test_existing_connection(nodedoc_zi_json, mock_connection):
    mock_connection.listNodesJSON.return_value = nodedoc_zi_json
    mock_connection.getString.return_value = "DataServer"

    Session("localhost", connection=mock_connection)
    mock_connection.assert_not_called()
    mock_connection.getString.assert_called_once_with("/zi/about/dataserver")
    mock_connection.listNodesJSON.assert_called_once_with("/zi/*")

    Session("localhost", connection=mock_connection, hf2=False)
    with pytest.raises(RuntimeError) as e_info:
        Session("localhost", connection=mock_connection, hf2=True)


def test_existing_connection_hf2(mock_connection):
    mock_connection.getString.return_value = "HF2DataServer"

    Session("localhost", connection=mock_connection)
    mock_connection.assert_not_called()
    mock_connection.getString.assert_called_with("/zi/about/dataserver")
    mock_connection.listNodesJSON.assert_not_called()

    Session("localhost", connection=mock_connection, hf2=True)
    with pytest.raises(RuntimeError) as e_info:
        Session("localhost", connection=mock_connection, hf2=False)


def test_wrong_port(mock_connection):
    def setup_side_effect(host, port, level):
        if level > 1:
            raise RuntimeError("Unsupported API level")
        return mock_connection.return_value

    mock_connection.side_effect = setup_side_effect
    # Test HF2 server
    mock_connection.return_value.getString.return_value = "HF2DataServer"
    Session("localhost", 8005)
    mock_connection.assert_called_with("localhost", 8005, 1)

    Session("localhost", 8005, hf2=True)
    with pytest.raises(RuntimeError) as e_info:
        Session("localhost", 8005, hf2=False)


def test_connect_device(
    zi_devices_json, mock_connection, session, nodedoc_dev1234_json
):

    connected_devices = ""

    def get_string_side_effect(arg):
        if arg == "/zi/devices":
            return zi_devices_json
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    def connect_device_side_effect(serial, _):
        nonlocal connected_devices
        devices = [key for key in json.loads(zi_devices_json).keys()]
        if serial.upper() not in devices:
            raise RuntimeError("device not visible to server")
        if serial not in connected_devices:
            if not connected_devices:
                connected_devices = serial
            else:
                connected_devices = connected_devices + "," + serial

    mock_connection.return_value.connectDevice.side_effect = connect_device_side_effect
    mock_connection.return_value.listNodesJSON.return_value = nodedoc_dev1234_json

    # device not visible
    with pytest.raises(KeyError) as e_info:
        session.connect_device("dev1111")
    assert "dev1111" in e_info.value.args[0].lower()

    # device visible
    device = session.connect_device("dev1234")
    mock_connection.return_value.connectDevice.assert_called_with("dev1234", "1GbE")
    assert device.serial == "dev1234"
    assert device.device_type == "Test"
    assert device.__class__.__name__ == "BaseInstrument"

    # access connected device
    assert "dev1234" in session.devices
    assert device == session.devices["dev1234"]
    assert len(session.devices) == 1
    for dev in session.devices:
        assert dev == "dev1234"

    # delete existing device
    # connot be overwritten
    with pytest.raises(LookupError) as e_info:
        session.devices["dev1234"] = None

    session.disconnect_device("dev1234")
    # Can be called as often as wanted
    session.disconnect_device("dev1234")


def test_connect_device_h2(
    data_dir, mock_connection, hf2_session, nodedoc_dev1234_json
):
    def connect_device_side_effect(serial, _):
        if serial == "dev1111":
            raise RuntimeError("dev1111 not visible")

    mock_connection.return_value.connectDevice.side_effect = connect_device_side_effect

    def get_string_side_effect(arg):
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    mock_connection.return_value.listNodesJSON.return_value = nodedoc_dev1234_json

    # device not visible
    with pytest.raises(RuntimeError) as e_info:
        hf2_session.connect_device("dev1111")
    assert "dev1111" in e_info.value.args[0].lower()

    # device visible
    device = hf2_session.connect_device("dev1234")
    mock_connection.return_value.connectDevice.assert_called_with("dev1234", "USB")
    assert device.serial == "dev1234"
    assert device.device_type == "Test"
    assert device.__class__.__name__ == "BaseInstrument"

    # access connected device
    assert "dev1234" in hf2_session.devices
    assert device == hf2_session.devices["dev1234"]
    assert len(hf2_session.devices) == 1
    for dev in hf2_session.devices:
        assert dev == "dev1234"

    with pytest.raises(RuntimeError) as e_info:
        hf2_session.devices.add_hf2_device("dev1234")

    # connect non HF2 device
    with patch("zhinst.toolkit.session.core.ziDiscovery", autospec=True) as discovery:
        discovery.return_value.get.return_value = {"devicetype": "Test234"}
        with pytest.raises(RuntimeError) as e_info:
            hf2_session.connect_device("dev5678")
    assert "dev5678" in e_info.value.args[0].lower()
    assert "test234" in e_info.value.args[0].lower()


def test_devices_visible(mock_connection, session, hf2_session):
    session.devices.visible()
    mock_connection.return_value.getString.assert_called_with("/zi/devices/visible")

    def get_string_side_effect(arg):
        if arg == "/zi/devices/visible":
            raise RuntimeError()

    mock_connection.return_value.getString.side_effect = get_string_side_effect
    with patch("zhinst.toolkit.session.core.ziDiscovery", autospec=True) as discovery:
        hf2_session.devices.visible()
        discovery.return_value.findAll.assert_called_once()


def test_sync(mock_connection, session):
    session.sync()
    mock_connection.return_value.sync.assert_called_once()


def test_poll(mock_connection, session):
    # default call
    # empty result
    result = session.poll()
    mock_connection.return_value.poll.assert_called_with(0.1, 500, flags=0, flat=True)
    assert not result

    # with result
    # TODO


def test_modules_repr(session):
    assert repr(session.modules) == "LabOneModules(localhost:8004)"


def test_raw_path_to_node(data_dir, mock_connection, session, zi_devices_json):
    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices":
            return zi_devices_json
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect
    result = session.raw_path_to_node("/dev1234/test/a/b/c")
    assert result.raw_tree == ("test", "a", "b", "c")
    assert result.root.prefix_hide == "dev1234"

    # Only absolut paths are supported
    with pytest.raises(RuntimeError) as e_info:
        session.raw_path_to_node("dev1234/test/a/b/c")

    # Not connected device
    with pytest.raises(RuntimeError) as e_info:
        session.raw_path_to_node("/dev1235/test/a/b/c")

    # with optional module
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.dataAcquisitionModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    result = session.raw_path_to_node("/device", module=session.modules.daq)
    assert result.raw_tree == ("device",)
    assert result.root.connection == session.modules.daq.raw_module

    # zi node
    result = session.raw_path_to_node(
        "/zi/about/commit",
    )


def test_awg_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_awg_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.awgModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    awg_module = session.modules.awg
    assert awg_module == session.modules.awg
    mock_connection.return_value.awgModule.assert_called_once()
    assert isinstance(awg_module, BaseModule)
    assert isinstance(awg_module.device, Node)


def test_daq_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.dataAcquisitionModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    daq_module = session.modules.daq
    assert daq_module == session.modules.daq
    mock_connection.return_value.dataAcquisitionModule.assert_called_once()
    assert isinstance(daq_module, DAQModule)
    assert isinstance(daq_module.device, Node)


def test_device_settings_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.deviceSettings.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    device_settings_module = session.modules.device_settings
    assert device_settings_module == session.modules.device_settings
    mock_connection.return_value.deviceSettings.assert_called_once()
    assert isinstance(device_settings_module, BaseModule)
    assert isinstance(device_settings_module.device, Node)


def test_impedance_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.impedanceModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    impedance_module = session.modules.impedance
    assert impedance_module == session.modules.impedance
    mock_connection.return_value.impedanceModule.assert_called_once()
    assert isinstance(impedance_module, BaseModule)
    assert isinstance(impedance_module.device, Node)


def test_mds_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.multiDeviceSyncModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    mds_module = session.modules.mds
    assert mds_module == session.modules.mds
    mock_connection.return_value.multiDeviceSyncModule.assert_called_once()
    assert isinstance(mds_module, BaseModule)
    assert isinstance(mds_module.device, Node)


def test_pid_advisor_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.pidAdvisor.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    pid_advisor_module = session.modules.pid_advisor
    assert pid_advisor_module == session.modules.pid_advisor
    mock_connection.return_value.pidAdvisor.assert_called_once()
    assert isinstance(pid_advisor_module, BaseModule)
    assert isinstance(pid_advisor_module.device, Node)


def test_precompensation_advisor_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.precompensationAdvisor.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    precompensation_advisor_module = session.modules.precompensation_advisor
    assert precompensation_advisor_module == session.modules.precompensation_advisor
    mock_connection.return_value.precompensationAdvisor.assert_called_once()
    assert isinstance(precompensation_advisor_module, BaseModule)
    assert isinstance(precompensation_advisor_module.device, Node)


def test_qa_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.quantumAnalyzerModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    qa_module = session.modules.qa
    assert qa_module == session.modules.qa
    mock_connection.return_value.quantumAnalyzerModule.assert_called_once()
    assert isinstance(qa_module, BaseModule)
    assert isinstance(qa_module.device, Node)


def test_scope_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.scopeModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    scope_module = session.modules.scope
    assert scope_module == session.modules.scope
    mock_connection.return_value.scopeModule.assert_called_once()
    assert isinstance(scope_module, BaseModule)
    assert isinstance(scope_module.device, Node)


def test_sweeper_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_sweeper_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.sweep.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    sweeper_module = session.modules.sweeper
    assert sweeper_module == session.modules.sweeper
    mock_connection.return_value.sweep.assert_called_once()
    assert isinstance(sweeper_module, SweeperModule)
    assert isinstance(sweeper_module.device, Node)


def test_shfqa_sweeper(session, mock_sweeper_daq):
    sweeper = session.modules.shfqa_sweeper
    assert sweeper == session.modules.shfqa_sweeper
    assert isinstance(sweeper, SHFQASweeper)


def test_session_wide_transaction(
    mock_connection, nodedoc_dev1234_json, session, shfqa, shfsg
):

    # Hack devices into the created once
    session._devices._devices = {"dev1": shfqa, "dev2": shfsg}

    with session.set_transaction():
        session.mds.groups[0].keepalive(1)
        shfqa.qachannels[0].centerfreq(100)
        shfsg.sgchannels[1].awg.time(100)
        session.mds.groups[0].keepalive(4)

    mock_connection.return_value.set.assert_called_once_with(
        [
            ("/zi/mds/groups/0/keepalive", 1),
            ("/dev1234/qachannels/0/centerfreq", 1000000000.0),  # lower bound ;)
            ("/dev1234/sgchannels/1/awg/time", 100),
            ("/zi/mds/groups/0/keepalive", 4),
        ]
    )

    # impossible to create two transactions
    with pytest.raises(RuntimeError) as e_info:
        with session.set_transaction():
            with session.set_transaction():
                session.mds.groups[0].keepalive(1)

    # add new device during transaction
    mock_connection.return_value.getString.side_effect = ["dev1234", "MFLI", ""]
    mock_connection.return_value.listNodesJSON.return_value = nodedoc_dev1234_json
    with session.set_transaction():
        session.mds.groups[0].keepalive(1)
        shfqa.qachannels[0].centerfreq(100)
        session.devices["dev1234"].demods[0].enable(1)
    mock_connection.return_value.set.assert_called_with(
        [
            ("/zi/mds/groups/0/keepalive", 1),
            ("/dev1234/qachannels/0/centerfreq", 1000000000.0),  # lower bound ;)
            ("/dev1234/demods/0/enable", 1),
        ]
    )



"""High-level generic lazy node tree for the zhinst.core package.

All interactions with an Zurich Instruments device or a LabOne
module happens through manipulating nodes. The
:class:`zhinst.toolkit.nodetree.nodetree.NodeTree` provides a pythonic way for
that.

>>> nodetree = NodeTree(connection)
>>> nodetree.example.nodes[8].test
    /example/nodes/8/test
"""

from zhinst.toolkit.nodetree.node import Node
from zhinst.toolkit.nodetree.nodetree import NodeTree


__all__ = ["Node", "NodeTree"]



"""Predefined parsers for device specific nodes."""
import logging

UHFQA_SAMPLE_RATE = 1.8e9
SHFQA_SAMPLE_RATE = 2e9
logger = logging.getLogger(__name__)


class Parse:
    """Input and output parsers for node parameters to validate and parse values."""

    @staticmethod
    def from_bool(value: bool) -> int:
        """Convert a boolean value to a integer value.

        Args:
            value: A boolean value.

        Returns:
            Integer value.
        """
        return int(value)

    @staticmethod
    def to_bool(value: int) -> bool:
        """Convert a integer value to a boolean value.

        Args:
            value: A integer value.

        Returns:
            Boolean value.
        """
        return bool(value)

    @staticmethod
    def phase(raw_phase: float) -> float:
        """Corrects the phase to -180 <= value <= 180.

        Args:
            raw_phase: Raw input phase.

        Returns:
            Corrected phase.
        """
        return (raw_phase + 180) % 360 - 180

    @staticmethod
    def greater_equal(value: float, limit: float) -> float:
        """Ensures that the value is greater or equal a lower limit.

        Args:
            value: Used value.
            limit: Minimum value returned.

        Returns:
            Clamped value.
        """
        if value < limit:
            logger.warning(
                f"The value {value:.3e} must be greater than or equal to "
                f"{limit:.3e} and will be rounded up to: "
                f"{limit:.3e}"
            )
            return limit
        else:
            return value

    @staticmethod
    def smaller_equal(value: float, limit: float) -> float:
        """Ensures that the value is smaller or equal a upper limit.

        Args:
            value: Used value.
            limit: Maximum value returned.

        Returns:
            Clamped value.
        """
        if value > limit:
            logger.warning(
                f"The value {value:.3e} must be smaller than or equal to "
                f"{limit:.3e} and will be rounded down to: "
                f"{limit:.3e}",
            )
            return limit
        else:
            return value

    @staticmethod
    def multiple_of(value: float, factor: float, rounding: str) -> float:
        """Rounds a value to a multiple of a given factor.

        Args:
            value: Input value.
            factor: Factor that the value needs to be multiple of.
            rounding: Method of rounding (nearest, down).

        Returns:
            Rounded value.
        """
        if abs(round(value / factor) * factor - value) < 1e-12:
            return value
        elif rounding == "nearest":
            v_rounded = round(value / factor) * factor
            logger.warning(
                f"The value {value:.3e} is not a multiple of "
                f"{factor:.3e} and will be rounded to nearest "
                f"multiple: {v_rounded:.3e}",
            )
            return v_rounded
        elif rounding == "down":
            v_rounded = round(value // factor) * factor
            logger.warning(
                f"The value {value:.3e} is not a multiple of "
                f"{factor:.3e} and will be rounded down to greatest "
                f"multiple: {v_rounded:.3e}",
            )
            return v_rounded
        raise RuntimeError(
            f"Invalid rounding type {rounding} only the "
            "following values are allowed: [nearest,down]"
        )


node_parser = {
    "SHFQA": {
        "scopes/0/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "scopes/0/channels/*/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "scopes/0/trigger/delay": {
            "SetParser": lambda v: Parse.multiple_of(v, 2e-9, "nearest"),
        },
        "scopes/0/length": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, 16),
                lambda v: Parse.smaller_equal(v, 2**18),
                lambda v: Parse.multiple_of(v, 16, "down"),
            ],
        },
        "scopes/0/segments/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "scopes/0/segments/count": {
            "SetParser": lambda v: Parse.greater_equal(v, 0),
        },
        "scopes/0/averaging/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "scopes/0/averaging/count": {
            "SetParser": lambda v: Parse.greater_equal(v, 0),
        },
        "qachannels/*/input/on": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "qachannels/*/output/on": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "qachannels/*/input/range": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, -50),
                lambda v: Parse.smaller_equal(v, 10),
                lambda v: Parse.multiple_of(v, 5, "nearest"),
            ],
        },
        "qachannels/*/output/range": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, -50),
                lambda v: Parse.smaller_equal(v, 10),
                lambda v: Parse.multiple_of(v, 5, "nearest"),
            ],
        },
        "qachannels/*/centerfreq": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, 1e9),
                lambda v: Parse.smaller_equal(v, 8e9),
                lambda v: Parse.multiple_of(v, 100e6, "nearest"),
            ],
        },
        "qachannels/*/generator/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "qachannels/*/generator/delay": {
            "SetParser": lambda v: Parse.multiple_of(v, 2e-9, "nearest"),
        },
        "qachannels/*/generator/single": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "qachannels/*/readout/result/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "qachannels/*/oscs/0/gain": {
            "SetParser": [
                lambda v: Parse.smaller_equal(v, 1.0),
                lambda v: Parse.greater_equal(v, 0.0),
            ],
        },
        "qachannels/*/oscs/0/freq": {
            "SetParser": [
                lambda v: Parse.smaller_equal(v, 500e6),
                lambda v: Parse.greater_equal(v, -500e6),
            ],
        },
        "qachannels/*/spectroscopy/length": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, 4),
                lambda v: Parse.smaller_equal(v, ((2**23) - 1) * 4),
                lambda v: Parse.multiple_of(v, 4, "down"),
            ],
        },
        "qachannels/*/spectroscopy/delay": {
            "SetParser": lambda v: Parse.multiple_of(v, 2e-9, "nearest")
        },
    },
    "SHFSG": {
        "system/clocks/referenceclock/out/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "system/clocks/referenceclock/out/freq": {
            "SetParser": lambda v: Parse.greater_equal(v, 0),
        },
        "sgchannels/*/centerfreq": {
            "SetParser": lambda v: Parse.greater_equal(v, 0),
        },
        "sgchannels/*/output/on": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "sgchannels/*/output/range": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, -30),
                lambda v: Parse.smaller_equal(v, 10),
                lambda v: Parse.multiple_of(v, 5, "nearest"),
            ],
        },
        "sgchannels/*/awg/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "sgchannels/*/awg/single": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "sgchannels/*/awg/outputs/*/enables/*": {
            "GetParser": Parse.to_bool,
        },
        "sgchannels/*/awg/outputs/*/gains/*": {
            "SetParser": [
                lambda v: Parse.smaller_equal(v, 1.0),
                lambda v: Parse.greater_equal(v, -1.0),
            ],
        },
        "sgchannels/*/oscs/*/freq": {
            "SetParser": [
                lambda v: Parse.smaller_equal(v, 1e9),
                lambda v: Parse.greater_equal(v, -1e9),
            ],
        },
        "sgchannels/*/sines/*/phaseshift": {
            "SetParser": Parse.phase,
        },
        "sgchannels/*/sines/*/oscselect": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, 0),
                lambda v: Parse.smaller_equal(v, 7),
                lambda v: Parse.multiple_of(v, 1, "nearest"),
            ],
        },
        "sgchannels/*/sines/*/harmonic": {
            "SetParser": [
                lambda v: Parse.greater_equal(v, 1),
                lambda v: Parse.smaller_equal(v, 1023),
                lambda v: Parse.multiple_of(v, 1, "nearest"),
            ],
        },
        "sgchannels/*/sines/*/i/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
        "sgchannels/*/sines/*/q/enable": {
            "GetParser": Parse.to_bool,
            "SetParser": Parse.from_bool,
        },
    },
}



"""Interface Enums for the zhinst.toolkit."""
from enum import Enum, IntEnum


class SHFQAChannelMode(Enum):
    """SHFQA channel mode."""

    SPECTROSCOPY = "spectroscopy"
    READOUT = "readout"


class AveragingMode(IntEnum):
    """Averaging modes.

    CYCLIC:
        All frequency points are measured once from start frequency to stop
        frequency. The sweeper then moves back to start frequency and repeats the
        sweep the number of times specified by the number of averages setting.

    SEQUENTIAL:
        A frequency point is measured the number of times specified by the
        number of averages setting. In other words, the same frequency point is
        measured repeatedly until the number of averages is reached and the
        sweeper then moves to the next frequency point.
    """

    CYCLIC = 0
    SEQUENTIAL = 1



import numpy as np
import pytest

from zhinst.toolkit.driver.modules.daq_module import DAQModule


@pytest.fixture()
def daq_module(data_dir, mock_connection, session):
    json_path = data_dir / "nodedoc_daq_test.json"
    with json_path.open("r", encoding="UTF-8") as file:
        nodes_json = file.read()
    mock_connection.return_value.dataAcquisitionModule.return_value.listNodesJSON.return_value = (
        nodes_json
    )
    yield DAQModule(mock_connection.return_value.dataAcquisitionModule(), session)


def test_repr(daq_module):
    assert "MagicMock(DataServerSession(localhost:8004))" in repr(daq_module)


def test_triggernode(daq_module, mock_connection, session):
    module_mock = mock_connection.return_value.dataAcquisitionModule.return_value
    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    # get
    module_mock.getString.return_value = "/dev1234/sample.a.b"
    result = daq_module.triggernode()
    assert result.raw_tree == ("sample", "a", "b")
    assert result.root.prefix_hide == "dev1234"

    module_mock.getString.return_value = "/dev1234/test/a/b"
    result = daq_module.triggernode()
    assert result.raw_tree == ("test", "a", "b")
    assert result.root.prefix_hide == "dev1234"

    module_mock.getString.return_value = "/dev1111/test/a/b"
    result = daq_module.triggernode()
    assert result == "/dev1111/test/a/b"

    # set
    daq_module.triggernode(session.devices["dev1234"].test.a.c)
    module_mock.set.assert_called_with("/triggernode", "/dev1234/test/a/c")

    daq_module.triggernode(session.devices["dev1234"].sample.a.c)
    module_mock.set.assert_called_with("/triggernode", "/dev1234/sample.a.c")

    daq_module.triggernode("dev123/test/hello")
    module_mock.set.assert_called_with("/triggernode", "dev123/test/hello")


def test_read(daq_module, mock_connection, session):
    connected_devices = "dev1234"

    def get_string_side_effect(arg):
        if arg == "/zi/devices/connected":
            return connected_devices
        if arg == "/dev1234/features/devtype":
            return "Test"
        raise RuntimeError("ZIAPINotFoundException")

    mock_connection.return_value.getString.side_effect = get_string_side_effect

    module_mock = mock_connection.return_value.dataAcquisitionModule.return_value
    module_mock.read.return_value = {
        "/awgcontrol": np.array([0]),
        "/device": ["1234"],
        "/dev1234/demods/0/sample.x.avg": [
            {
                "header": {"systemtime": np.array([1643040922025522])},
                "timestamp": np.array([[1, 2, 3, 4, 5]]),
                "value": np.array([[1, 2, 3, 4, 5]]),
            }
        ],
    }
    result = daq_module.read()
    assert len(result) == 3
    result = result[session.devices["dev1234"].demods[0].sample.x.avg]
    assert "systemtime" in result[0].header
    assert (result[0].time == np.array([[0, 1, 2, 3, 4]]) / 60e6).all()
    assert (result[0].value == np.array([[1, 2, 3, 4, 5]])).all()
    assert result[0].frequency == None
    assert result[0].shape == (1, 5)

    result = daq_module.read(raw=True)
    assert len(result) == 3
    result = result[session.devices["dev1234"].demods[0].sample.x.avg]
    assert "systemtime" in result[0]["header"]
    assert (result[0]["timestamp"] == np.array([[1, 2, 3, 4, 5]])).all()
    assert (result[0]["value"] == np.array([[1, 2, 3, 4, 5]])).all()

    module_mock.read.return_value = {
        "/awgcontrol": np.array([0]),
        "/device": ["1234"],
        "/dev1234/demods/0/sample.x.avg": [
            {
                "header": {"systemtime": np.array([1643040922025522])},
                "timestamp": np.array([[1, 2, 3, 4, 5]]),
                "value": np.array([[1, 2, 3, 4, 5]]),
            },
            {
                "header": {"systemtime": np.array([1643040922040000])},
                "timestamp": np.array([[6, 7, 8, 9, 10]]),
                "value": np.array([[5, 4, 3, 2, 1]]),
            },
        ],
        "/dev1234/demods/1/sample.xiy.fft": [
            {
                "header": {"gridcoldelta": np.array([0.00059733])},
                "timestamp": np.array([[1, 2, 3, 4, 5]]),
                "value": np.array([[9, 9, 9, 9, 9]]),
            },
        ],
    }
    result = daq_module.read()
    assert len(result) == 4
    result0 = result[session.devices["dev1234"].demods[0].sample.x.avg]
    result1 = result[session.devices["dev1234"].demods[1].sample.xiy.fft]
    assert "systemtime" in result0[0].header
    assert (result0[0].time == np.array([[0, 1, 2, 3, 4]]) / 60e6).all()
    assert (result0[0].value == np.array([[1, 2, 3, 4, 5]])).all()
    assert result0[0].frequency == None
    assert result0[0].shape == (1, 5)
    assert (result0[1].time == np.array([[0, 1, 2, 3, 4]]) / 60e6).all()
    assert (result0[1].value == np.array([[5, 4, 3, 2, 1]])).all()
    assert result0[1].frequency == None
    assert result0[1].shape == (1, 5)
    assert "gridcoldelta" in result1[0].header
    assert np.allclose(
        result1[0].frequency,
        np.array([[-1.19466e-03, -5.9733e-04, -5.42101e-20, 5.9733e-04, 1.19466e-03]]),
    )
    assert (result1[0].value == np.array([[9, 9, 9, 9, 9]])).all()
    assert result1[0].time == None
    assert result1[0].shape == (1, 5)


def test_subscribe(daq_module, mock_connection):
    module_mock = mock_connection.return_value.dataAcquisitionModule.return_value
    daq_module.subscribe(daq_module.test.sample.x.avg)
    module_mock.subscribe.assert_called_with("/test/sample.x.avg")


def test_unsubscribe(daq_module, mock_connection):
    module_mock = mock_connection.return_value.dataAcquisitionModule.return_value
    daq_module.unsubscribe(daq_module.test.sample.x.avg)
    module_mock.unsubscribe.assert_called_with("/test/sample.x.avg")


def test_execute(daq_module, mock_connection):
    module_mock = mock_connection.return_value.dataAcquisitionModule.return_value
    daq_module.execute()
    module_mock.execute.assert_called_with()



"""SHFSG Instrument Driver."""

import logging
import typing as t
import warnings

import zhinst.utils.shfsg as utils

from zhinst.toolkit.driver.devices.base import BaseInstrument
from zhinst.toolkit.driver.nodes.awg import AWG
from zhinst.toolkit.nodetree import Node
from zhinst.toolkit.nodetree.helper import lazy_property
from zhinst.toolkit.nodetree.node import NodeList

logger = logging.getLogger(__name__)

if t.TYPE_CHECKING:
    from zhinst.toolkit.session import Session


class AWGCore(AWG):
    """AWG Core Node."""

    def configure_marker_and_trigger(
        self,
        *,
        trigger_in_source: str,
        trigger_in_slope: str,
        marker_out_source: str,
    ) -> None:
        """Configures the trigger inputs and marker outputs of the AWG.

        Args:
            trigger_in_source: Alias for the trigger input used by the
                sequencer. For a list of available values use:
                `available_trigger_inputs`
            trigger_in_slope: Alias for the slope of the input trigger
                used by sequencer. For a list of available values use
                `available_trigger_inputs`
            marker_out_source: Alias for the marker output source used by
                the sequencer. For a list of available values use
                `available_trigger_slopes`
        """
        utils.configure_marker_and_trigger(
            self._daq_server,
            self._serial,
            self._index,
            trigger_in_source=trigger_in_source,
            trigger_in_slope=trigger_in_slope,
            marker_out_source=marker_out_source,
        )

    @property
    def available_trigger_inputs(self) -> t.List[str]:
        """List the available trigger sources for the sequencer."""
        return [
            option.enum
            for option in self.auxtriggers[0].channel.node_info.options.values()
        ]

    @property
    def available_trigger_slopes(self) -> t.List[str]:
        """List the available trigger slopes for the sequencer."""
        return [
            option.enum
            for option in self.auxtriggers[0].slope.node_info.options.values()
        ]

    @property
    def available_marker_outputs(self) -> t.List[str]:
        """List the available trigger marker outputs for the sequencer."""
        return [
            option.enum
            for option in self.root.sgchannels[
                self._index
            ].marker.source.node_info.options.values()
        ]


class SGChannel(Node):
    """Signal Generator Channel for the SHFSG.

    :class:`SGChannel` implements basic functionality to configure SGChannel
    settings of the :class:`SHFSG` instrument.

    Args:
        device: SHFQA device object.
        session: Underlying session.
        tree: Node tree (node path as tuple) of the corresponding node.
    """

    def __init__(
        self,
        device: "SHFSG",
        session: "Session",
        tree: t.Tuple[str, ...],
    ):
        super().__init__(device.root, tree)
        self._index = int(tree[-1])
        self._device = device
        self._serial = device.serial
        self._session = session

    def configure_channel(
        self,
        *,
        enable: bool,
        output_range: int,
        center_frequency: float,
        rf_path: bool,
    ) -> None:
        """Configures the RF input and output.

        Args:
            enable: Flag if the signal output should be enabled.
            output_range: Maximal range of the signal output power in dBm
            center_frequency: Center frequency before modulation
            rf_path: Flag if the RF(True) or LF(False) path should be
                configured.
        """
        utils.configure_channel(
            self._session.daq_server,
            self._device.serial,
            self._index,
            enable=int(enable),
            output_range=output_range,
            center_frequency=center_frequency,
            rflf_path=int(rf_path),
        )

    def configure_pulse_modulation(
        self,
        *,
        enable: bool,
        osc_index: int = 0,
        osc_frequency: float = 100e6,
        phase: float = 0.0,
        global_amp: float = 0.5,
        gains: tuple = (1.0, -1.0, 1.0, 1.0),
        sine_generator_index: int = 0,
    ) -> None:
        """Configure the pulse modulation.

        Configures the sine generator to digitally modulate the AWG output, for
        generating single sideband AWG signals

        Args:
            enable: Flag if the modulation should be enabled.
            osc_index: Selects which oscillator to use
            osc_frequency: Oscillator frequency used to modulate the AWG
                outputs. (default = 100e6)
            phase: Sets the oscillator phase. (default = 0.0)
            global_amp: Global scale factor for the AWG outputs. (default = 0.5)
            gains: Sets the four amplitudes used for single sideband generation.
                Default values correspond to upper sideband with a positive
                oscillator frequency. (default = (1.0, -1.0, 1.0, 1.0))
            sine_generator_index: Selects which sine generator to use on a
                given channel.
        """
        utils.configure_pulse_modulation(
            self._session.daq_server,
            self._device.serial,
            self._index,
            enable=int(enable),
            osc_index=osc_index,
            osc_frequency=osc_frequency,
            phase=phase,
            global_amp=global_amp,
            gains=gains,
            sine_generator_index=sine_generator_index,
        )

    def configure_sine_generation(
        self,
        *,
        enable: bool,
        osc_index: int = 0,
        osc_frequency: float = 100e6,
        phase: float = 0.0,
        gains: tuple = (0.0, 1.0, 1.0, 0.0),
        sine_generator_index: int = 0,
    ) -> None:
        """Configures the sine generator output.

        Configures the sine generator output of a specified channel for generating
        continuous wave signals without the AWG.

        Args:
            enable: Flag if the sine generator output should be enabled.
            osc_index: Selects which oscillator to use
            osc_frequency: Oscillator frequency used by the sine generator
                (default = 100e6)
            phase: Sets the oscillator phase. (default = 0.0)
            gains: Sets the four amplitudes used for single sideband
                generation. Default values correspond to upper sideband with a
                positive oscillator frequency.
                Gains are set in the following order I/sin, I/cos, Q/sin, Q/cos.
                (default = (0.0, 1.0, 1.0, 0.0))
            sine_generator_index: Selects which sine generator to use on a given
                channel
        """
        utils.configure_sine_generation(
            self._session.daq_server,
            self._device.serial,
            self._index,
            enable=int(enable),
            osc_index=osc_index,
            osc_frequency=osc_frequency,
            phase=phase,
            gains=gains,
            sine_generator_index=sine_generator_index,
        )

    @property
    def awg_modulation_freq(self) -> float:
        """Modulation frequency of the AWG.

        Depends on the selected oscillator.
        """
        selected_osc = self.sines[0].oscselect()
        return self.oscs[selected_osc].freq()

    @lazy_property
    def awg(self) -> AWGCore:
        """AWG."""
        return AWGCore(
            self._root,
            self._tree + ("awg",),
            self._device.serial,
            self._index,
            self._device.device_type,
            self._device.device_options,
        )


class SHFSG(BaseInstrument):
    """High-level driver for the Zurich Instruments SHFSG."""

    def factory_reset(self, *, deep: bool = True) -> None:
        """Load the factory default settings.

        Args:
            deep: A flag that specifies if a synchronization
                should be performed between the device and the data
                server after loading the factory preset (default: True).
        """
        warnings.warn("Factory preset is not yet supported for SHFSG.", RuntimeWarning)
        logger.warning("Factory preset is not yet supported in SHFSG.")

    @lazy_property
    def sgchannels(self) -> t.Sequence[SGChannel]:
        """A Sequence of SG Channels."""
        return NodeList(
            [
                SGChannel(self, self._session, self._tree + ("sgchannels", str(i)))
                for i in range(len(self["sgchannels"]))
            ],
            self._root,
            self._tree + ("sgchannels",),
        )



"""Toolkit adaption for the zhinst.utils.SHFSweeper."""

import json
import logging
import re
import typing as t
from collections import OrderedDict
from dataclasses import asdict
from enum import IntEnum
from pathlib import Path

import numpy as np
from zhinst.utils.shf_sweeper import AvgConfig, EnvelopeConfig, RfConfig
from zhinst.utils.shf_sweeper import ShfSweeper as CoreSweeper
from zhinst.utils.shf_sweeper import SweepConfig, TriggerConfig
from zhinst.core import ziDAQServer

from zhinst.toolkit.driver.parsers import Parse
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.connection_dict import ConnectionDict
from zhinst.toolkit.nodetree.helper import NodeDoc

if t.TYPE_CHECKING:
    from zhinst.toolkit.driver.devices import DeviceType
    from zhinst.toolkit.session import Session

logger = logging.getLogger(__name__)


class SHFQASweeper(Node):
    """Toolkit adaption for the zhinst.utils.SHFSweeper.

    For now the general sweeper module does not support the SHFQA. However a
    python based implementation called ``SHFSweeper`` does already provide
    this functionality. The ``SHFSweeper`` is part of the ``zhinst`` module
    and can be found in the utils.

    Toolkit wraps around the ``SHFSweeper`` and exposes a interface that is
    similar to the LabOne modules, meaning the parameters are exposed in a
    node tree like structure.

    All parameters can be accessed through their corresponding node:
    * device: Device to run the sweeper with
    * sweep: Frequency range settings for a sweep
    * rf: RF in- and output settings for a sweep
    * average: Averaging settings for a sweep
    * trigger: Settings for the trigger
    * envelope: Settings for defining a complex envelope for pulsed spectroscopy

    The underlying module is updated with the parameter changes automatically.
    Every functions from the underlying SHFSweeper module is exposed and can be
    used in the same way.

    Args:
        daq_server: Client Session that should be used by the sweeper.
        session: Session to the Data Server.
    """

    def __init__(self, session: "Session"):
        self._config_classes = {
            SweepConfig: ("sweep", "sweep_config"),
            RfConfig: ("rf", "rf_config"),
            AvgConfig: ("average", "avg_config"),
            TriggerConfig: ("trigger", "trig_config"),
            EnvelopeConfig: ("envelope", "envelope_config"),
        }
        self._renamed_nodes = {
            "imp50": "input_impedance",
            "use_sequencer": "mode",
            "force_sw_trigger": "sw_trigger_mode",
        }
        super().__init__(self._create_nodetree(), tuple())
        self._daq_server = ziDAQServer(
            session.daq_server.host,
            session.daq_server.port,
            6,
        )
        self._raw_module = CoreSweeper(self._daq_server, "")
        self._session = session
        self.root.update_nodes(
            {
                "/device": {
                    "GetParser": self._get_device,
                    "SetParser": self._set_device,
                },
                "/envelope/enable": {
                    "GetParser": Parse.to_bool,
                    "SetParser": Parse.from_bool,
                },
            },
            raise_for_invalid_node=False,
        )

    def __repr__(self):
        return str(f"SHFQASweeper({repr(self._session)})")

    def _get_device(self, serial: str) -> t.Union["DeviceType", str]:
        """Convert a device serial into a toolkit device object.

        Args:
            serial: Serial of the device

        Returns:
            Toolkit device object. If the serial does not
                match to a connected device the serial is returned instead.
        """
        try:
            return self._session.devices[serial]
        except (RuntimeError, KeyError):
            return serial

    def _set_device(self, value: t.Union["DeviceType", str]) -> str:
        """Convert a toolkit device object into a serial string.

        Args:
            value: Toolkit device object
                (can also be a serial string directly)

        Returns:
            Device serial
        """
        serial = ""
        try:
            serial = value.serial  # type: ignore
        except AttributeError:
            serial = value
        self._raw_module = CoreSweeper(self._daq_server, serial)
        return serial

    def _create_nodetree(self) -> NodeTree:
        """Create node tree for the SHFQA sweeper.

        Uses the hardcoded "resources/shfqa_sweeper_nodes.json" information
        and the SHFSweeper data classes to automatically create a valid node
        tree. (Automatically adds new parameters with dummy information)

        Returns:
            node tree for the shfqa sweeper
        """
        json_path = Path(__file__).parent / "../../resources/shfqa_sweeper_nodes.json"
        with json_path.open("r") as file:
            raw_info: NodeDoc = json.loads(file.read())
        values: t.Dict[str, t.Any] = {}
        info: NodeDoc = {}
        for config_class, parent_name in self._config_classes.items():
            for parameter, default_value in asdict(config_class()).items():
                node = (
                    f"/{parent_name[0]}/{self._renamed_nodes.get(parameter,parameter)}"
                )
                try:
                    info[node] = raw_info[node]
                except KeyError:
                    # node not in json
                    logger.warning(f"{node} is missing in {json_path}.")
                    type_mapping = {
                        int: "Integer (64 bit)",
                        float: "Double",
                        str: "String",
                        bool: "Integer (64 bit)",
                        np.ndarray: "ZIVectorData",
                    }
                    info[node] = {
                        "Node": node,
                        "Description": node,
                        "Properties": "Read, Write",
                        "Unit": "None",
                        "Type": type_mapping[type(default_value)],
                    }

                if "Options" in info[node]:
                    option_map = {}
                    for key, value in info[node]["Options"].items():
                        options = re.findall(r'"(.+?)"[,:]+', value)
                        option_map.update({x: int(key) for x in options})
                    values[node] = option_map.get(default_value, default_value)
                else:
                    values[node] = default_value
        info["/device"] = raw_info["/device"]
        values["/device"] = ""
        info["/envelope/enable"] = raw_info["/envelope/enable"]
        values["/envelope/enable"] = 0

        return NodeTree(ConnectionDict(values, info))

    def _update_settings(self) -> None:
        """Update the ShfSweeper settings from the node tree.

        Converts the nodetree into a valid configuration for the SHFSweeper.
        """
        if not self.device():
            raise RuntimeError(
                "The device serial needs to be set before using the module."
            )
        data = OrderedDict()
        for config_class, parent_name in self._config_classes.items():
            config = OrderedDict()
            for parameter in asdict(config_class()):
                value = self[parent_name[0]][
                    self._renamed_nodes.get(parameter, parameter)
                ]()
                if isinstance(value, IntEnum):
                    value = value.name
                config[parameter] = value
            data[parent_name[1]] = config_class(**config)

        # special treatment for the envelope config
        if not self.envelope.enable():
            data.pop("envelope_config")

        # special treatment for the imp50
        try:
            data["trig_config"].imp50 = data["trig_config"].imp50 == "imp50"
        except AttributeError:
            logger.warning(
                "imp50 setting is no longer available in the shf_sweeper class."
            )
        # special treatment for mode
        try:
            data["sweep_config"].use_sequencer = (
                data["sweep_config"].use_sequencer == "sequencer-based"
            )
        except AttributeError:
            logger.warning(
                "use_sequencer setting is no longer available in the shf_sweeper class."
            )
        # special treatment for trigger source
        try:
            data["trig_config"].source = (
                data["trig_config"].source
                if data["trig_config"].source != "auto"
                else None
            )
        except AttributeError:
            logger.warning(
                "source setting is no longer available in the shf_sweeper class."
            )
        # special treatment for the force_sw_trigger
        try:
            data["trig_config"].force_sw_trigger = (
                data["trig_config"].force_sw_trigger == "force"
            )
        except AttributeError:
            logger.warning(
                "force_sw_trigger setting is no longer available in the shf_sweeper."
            )
        self._raw_module.configure(**data)

    def run(self) -> dict:
        """Perform a sweep with the specified settings.

        This method wraps around the `run` method of
        `zhinst.utils.shf_sweeper`

        Returns:
             A dictionary with measurement data of the last sweep.
        """
        self._update_settings()
        return self._raw_module.run()

    def get_result(self) -> dict:
        """Get the measurement data of the last sweep.

        This method wraps around the `get_result` method of
        `zhinst.utils.shf_sweeper`

        Returns:
             A dictionary with measurement data of the last sweep.
        """
        self._update_settings()
        return self._raw_module.get_result()

    def plot(self) -> None:
        """Plot power over frequency for last sweep.

        This method wraps around the `plot` method of
        `zhinst.utils.shf_sweeper`
        """
        self._update_settings()
        return self._raw_module.plot()

    def get_offset_freq_vector(self) -> t.Any:
        """Get vector of frequency points.

        This method wraps around the `get_offset_freq_vector` method of
        `zhinst.utils.shf_sweeper`

        Returns:
            Vector of frequency points.
        """
        self._update_settings()
        return self._raw_module.get_offset_freq_vector()



"""UHFQA Instrument Driver."""

import typing as t

import numpy as np

from zhinst.toolkit.driver.devices import UHFLI
from zhinst.toolkit.nodetree import Node, NodeTree
from zhinst.toolkit.nodetree.helper import (
    create_or_append_set_transaction,
    lazy_property,
)
from zhinst.toolkit.nodetree.node import NodeList
from zhinst.toolkit.waveform import Waveforms

Numpy2DArray = t.TypeVar("Numpy2DArray")


class Integration(Node):
    """Integration part for the UHFQA.

    Args:
        root: Underlying node tree.
        tree: tree (node path as tuple) of the corresponding node.

    .. versionadded:: 0.3.2
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
    ):
        super().__init__(root, tree)

    def write_integration_weights(self, weights: t.Union[Waveforms, dict]) -> None:
        """Upload complex integration weights.

        The weight functions are applied to the real and imaginary part of
        the input signal. In the hardware the weights are implemented
        as 17-bit integers.

        Args:
            weights: Dictionary containing the weight functions, where
                keys correspond to the indices of the integration weights to be
                configured.

        Note:
            Does not raise an error when sample limit is exceeded, but applies only
            the maximum number of samples. Please refer to LabOne node documentation
            for the number of maximum integration weight samples.

        Note:
            This function calls both `/qas/n/integration/weights/n/real` and
            `/qas/n/integration/weights/n/imag` nodes.

            If only real or imaginary part is defined, the number of defined samples
            from the other one is zeroed.
        """
        waveform_dict = {}
        if isinstance(weights, Waveforms):
            for slot in weights.keys():
                waveform_dict[slot] = weights.get_raw_vector(slot, complex_output=True)
        else:
            waveform_dict = weights
        with create_or_append_set_transaction(self._root):
            for key, waveform in waveform_dict.items():
                self.weights[key].real(np.copy(waveform.real))
                self.weights[key].imag(np.copy(waveform.imag))


class QAS(Node):
    """Quantum Analyzer Channel for the UHFQA.

    Args:
        root: Underlying node tree.
        tree: tree (node path as tuple) of the corresponding node.
    """

    def __init__(
        self,
        root: NodeTree,
        tree: tuple,
    ):
        super().__init__(root, tree)

    def crosstalk_matrix(self, matrix: Numpy2DArray = None) -> t.Optional[Numpy2DArray]:
        """Sets or gets the crosstalk matrix of the UHFQA as a 2D array.

        Args:
            matrix: The 2D matrix used in the digital signal
                processing path to compensate for crosstalk between the
                different channels. The given matrix can also be a part
                of the entire 10 x 10 matrix. Its maximum dimensions
                are 10 x 10 (default: None).

        Returns:
            If no argument is given the method returns the current
            crosstalk matrix as a 2D numpy array.

        Raises:
            ValueError: If the matrix size exceeds the maximum size of
                10 x 10

        """
        if matrix is None:
            m = np.zeros((10, 10))
            for r in range(10):
                for c in range(10):
                    m[r, c] = self.crosstalk.rows[r].cols[c]()
            return m
        else:
            rows, cols = matrix.shape  # type: ignore[attr-defined]
            if rows > 10 or cols > 10:
                raise ValueError(
                    f"The shape of the given matrix is {rows} x {cols}. "
                    "The maximum size is 10 x 10."
                )
            for r in range(rows):
                for c in range(cols):
                    self.crosstalk.rows[r].cols[c](matrix[r, c])  # type: ignore[index]
        return None

    def adjusted_delay(self, value: int = None) -> int:
        """Set or get the adjustment in the quantum analyzer delay.

        Adjusts the delay that defines the time at which the integration starts
        in relation to the trigger signal of the weighted integration units.

        Depending if the deskew matrix is bypassed there exists a different
        default delay. This function can be used to add an additional delay to
        the default delay.

        Args:
            value: Number of additional samples to adjust the delay. If not
                specified this function will just return the additional delay
                currently set.

        Returns:
            The adjustment in delay in units of samples.

        Raises:
            ValueError: If the adjusted quantum analyzer delay is outside the
                allowed range of 1021 samples.

        """
        # The default delay value is defined by wether the deskew matrix is
        # bypassed or not.
        default_delay = 184 if self.bypass.deskew() else 200
        if value is None:
            return self.delay() - default_delay
        # Round down to greatest multiple of 4, as in LabOne.
        qa_delay_user = int(value // 4) * 4
        # Calculate final value of adjusted QA delay.
        qa_delay_adjusted = qa_delay_user + default_delay
        # Check if final delay is between 0 and 1020.
        if qa_delay_adjusted not in range(0, 1021):
            raise ValueError(
                "The quantum analyzer delay is out of range (0 <= "
                f"{qa_delay_user} + {default_delay} <= 1021)"
            )
        # Write the adjusted delay value to the node.
        self.delay(qa_delay_adjusted)
        return qa_delay_user

    @lazy_property
    def integration(self) -> Integration:
        """Integration.

        .. versionadded:: 0.3.2
        """
        return Integration(self.root, self._tree + ("integration",))


class UHFQA(UHFLI):
    """High-level driver for the Zurich Instruments UHFQA."""

    def enable_qccs_mode(self) -> None:
        """Configure the instrument to work with PQSC.

        This method sets the reference clock source and DIO settings
        correctly to connect the instrument to the PQSC.

        Info:
            Use ``factory_reset`` to reset the changes if necessary
        """
        with create_or_append_set_transaction(self._root):
            # Use external 10 MHz clock as reference
            self.system.extclk("external")
            # Configure DIO to be used in a QCCS
            # Clock DIO internally with a frequency of 50 MHz
            self.dios[0].extclk("internal")
            # Set DIO output values to QA results compatible with QCCS
            self.dios[0].mode("qa_result_qccs")
            # Drive the two least significant bytes of the DIO port
            self.dios[0].drive(0b0011)
            # Set correct DIO triggering in the AWG sequencer
            self.awgs[0].dio.strobe.slope("off")
            self.awgs[0].dio.valid.index(16)
            self.awgs[0].dio.valid.polarity("high")

    @lazy_property
    def qas(self) -> t.Sequence[QAS]:
        """A Sequence of QAS."""
        return NodeList(
            [
                QAS(self.root, self._tree + ("qas", str(i)))
                for i in range(len(self["qas"]))
            ],
            self._root,
            self._tree + ("qas",),
        )
